"""Contains custom components used inside the dashboard."""

from typing import Literal

import dash_bootstrap_components as dbc
import vizro.models as vm
from dash import dcc, html

try:
    from pydantic.v1 import Field
except ImportError:  # pragma: no cov
    from pydantic import Field


class CodeClipboard(vm.VizroBaseModel):
    """Code snippet with a copy to clipboard button."""

    type: Literal["code_clipboard"] = "code_clipboard"
    code: str
    language: str = ""

    def build(self):
        """Returns the code clipboard component inside an accordion."""
        markdown_code = "\n".join([f"```{self.language}", self.code, "```"])
        return dbc.Accordion(
            [
                dbc.AccordionItem(
                    html.Div(
                        [
                            dcc.Markdown(markdown_code, id=self.id),
                            dcc.Clipboard(target_id=self.id, className="code-clipboard"),
                        ],
                        className="code-clipboard-container",
                    ),
                    title="SHOW CODE",
                )
            ],
            start_collapsed=False,
        )


class Markdown(vm.VizroBaseModel):
    """Markdown component."""

    type: Literal["markdown"] = "markdown"
    text: str = Field(
        ..., description="Markdown string to create card title/text that should adhere to the CommonMark Spec."
    )
    classname: str = ""

    def build(self):
        """Returns a markdown component with an optional classname."""
        return dcc.Markdown(id=self.id, children=self.text, dangerously_allow_html=False, className=self.classname)


class FlexContainer(vm.Container):
    """Custom flex `Container`."""

    type: Literal["flex_container"] = "flex_container"
    title: str = None  # Title exists in vm.Container but we don't want to use it here.

    def build(self):
        """Returns a flex container."""
        return html.Div(
            id=self.id, children=[component.build() for component in self.components], className="flex-container"
        )


vm.Container.add_type("components", FlexContainer)
vm.Container.add_type("components", Markdown)
vm.Page.add_type("components", CodeClipboard)



import pandas as pd
from asserts import assert_component_equal
from dash import dash_table
from vizro.models.types import capture
from vizro.tables import dash_data_table

data = pd.DataFrame(
    {
        "cat": ["a", "b", "c"],
        "int": [4, 5, 6],
        "float": [7.3, 8.2, 9.1],
        "date": pd.to_datetime(["2021/01/01", "2021/01/02", "2021/01/03"]),
    }
)
columns = [
    {"id": "cat", "name": "cat"},
    {"id": "int", "name": "int"},
    {"id": "float", "name": "float"},
    {"id": "date", "name": "date"},
]
data_in_table = [
    {"cat": "a", "int": 4, "float": 7.3, "date": pd.Timestamp("2021-01-01 00:00:00")},
    {"cat": "b", "int": 5, "float": 8.2, "date": pd.Timestamp("2021-01-02 00:00:00")},
    {"cat": "c", "int": 6, "float": 9.1, "date": pd.Timestamp("2021-01-03 00:00:00")},
]


class TestDashDataTable:
    def test_dash_data_table(self):
        table = dash_data_table(data_frame=data)()
        assert_component_equal(
            table,
            dash_table.DataTable(
                columns=columns,
                data=data_in_table,
                style_as_list_view=True,
                style_cell={"position": "static"},
                style_data={"border_bottom": "1px solid var(--border-subtle-alpha-01)", "height": "40px"},
                style_header={
                    "border_bottom": "1px solid var(--state-overlays-selected-hover)",
                    "border_top": "1px solid var(--main-container-bg-color)",
                    "height": "32px",
                },
                style_data_conditional=[
                    {
                        "if": {"state": "active"},
                        "backgroundColor": "var(--state-overlays-selected)",
                        "border": "1px solid var(--state-overlays-selected)",
                    }
                ],
            ),
        )

    def test_custom_dash_data_table(self):
        @capture("table")
        def custom_dash_data_table(data_frame):
            return dash_table.DataTable(
                columns=[{"name": col, "id": col} for col in data_frame.columns],
                data=data_frame.to_dict("records"),
            )

        table = custom_dash_data_table(data_frame=data)()

        assert_component_equal(
            table,
            dash_table.DataTable(
                data=data_in_table,
                columns=columns,
            ),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal

import dash_mantine_components as dmc
from dash import html

try:
    from pydantic.v1 import validator
except ImportError:  # pragma: no cov
    from pydantic import validator

from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call, validate_min_length

if TYPE_CHECKING:
    from vizro.models._components import Container


class Tabs(VizroBaseModel):
    """Tabs to group together a set of containers on a page.

    Args:
        type (Literal["tabs"]): Defaults to `"tabs"`.
        tabs (List[Container]): See [`Container`][vizro.models.Container].

    """

    type: Literal["tabs"] = "tabs"
    tabs: List[Container]

    _validate_tabs = validator("tabs", allow_reuse=True, always=True)(validate_min_length)

    @_log_call
    def build(self):
        tabs_list = dmc.TabsList(
            [dmc.Tab(tab.title, value=tab.id, className="tab__title") for tab in self.tabs],
            className="tabs__list",
        )

        tabs_panels = [
            dmc.TabsPanel(html.Div([tab.build()], className="tab__content"), value=tab.id, className="tabs__panel")
            for tab in self.tabs
        ]

        return dmc.Tabs(
            id=self.id,
            value=self.tabs[0].id,
            children=[tabs_list, *tabs_panels],
            persistence=True,
            persistence_type="session",
            className="tabs",
        )



"""Contains custom charts used inside the dashboard."""

from typing import List

import pandas as pd
import vizro.plotly.express as px
from plotly import graph_objects as go
from plotly.subplots import make_subplots
from vizro.models.types import capture


# TODO: consider how this should be represented in the code example files. Since the code is copy and pasted
# it can get out of sync. But probably we don't want the docstrings in the short code snippet.
# Ultimately these charts will probably move to vizro.charts anyway.
@capture("graph")
def butterfly(data_frame: pd.DataFrame, x1: str, x2: str, y: str) -> go.Figure:
    """Creates a custom butterfly chart using Plotly's go.Figure.

    A butterfly chart is a type of bar chart where two sets of bars are displayed back-to-back, often used to compare
    two sets of data.

    Args:
        data_frame (pd.DataFrame): The data source for the chart.
        x1 (str): The name of the column in the data frame for the first set of bars (negative values).
        x2 (str): The name of the column in the data frame for the second set of bars (positive values).
        y (str): The name of the column in the data frame for the y-axis (categories).

    Returns:
        go.Figure: A Plotly Figure object representing the butterfly chart.

    """
    fig = go.Figure()
    fig.add_trace(
        go.Bar(
            x=-data_frame[x1],
            y=data_frame[y],
            orientation="h",
            name=x1,
        )
    )
    fig.add_trace(
        go.Bar(
            x=data_frame[x2],
            y=data_frame[y],
            orientation="h",
            name=x2,
        )
    )
    fig.update_layout(barmode="relative")
    return fig


@capture("graph")
def sankey(data_frame: pd.DataFrame, source: str, target: str, value: str, labels: List[str]) -> go.Figure:
    """Creates a custom sankey chart using Plotly's `go.Sankey`.

    A Sankey chart is a type of flow diagram where the width of the arrows is proportional to the flow rate.
    It is used to visualize the flow of resources or data between different stages or categories.

    Args:
        data_frame (pd.DataFrame): The data source for the chart.
        source (str): The name of the column in the data frame for the source nodes.
        target (str): The name of the column in the data frame for the target nodes.
        value (str): The name of the column in the data frame for the values representing the flow between nodes.
        labels (List[str]): A list of labels for the nodes.

    Returns:
        go.Figure: A Plotly Figure object representing the Sankey chart.

    For detailed information on additional parameters and customization, refer to the Plotly documentation:
    https://plotly.com/python/reference/sankey/

    """
    fig = go.Figure(
        data=[
            go.Sankey(
                node={
                    "pad": 16,
                    "thickness": 16,
                    "label": labels,
                },
                link={
                    "source": data_frame[source],
                    "target": data_frame[target],
                    "value": data_frame[value],
                    "label": labels,
                    "color": "rgba(205, 209, 228, 0.4)",
                },
            )
        ]
    )
    fig.update_layout(barmode="relative")
    return fig


@capture("graph")
def column_and_line(data_frame: pd.DataFrame, x: str, y_column: str, y_line: str) -> go.Figure:
    """Creates a combined column and line chart using Plotly.

    This function generates a chart with a bar graph for one variable (y-axis 1) and a line graph for another variable
    (y-axis 2), sharing the same x-axis. The y-axes for the bar and line graphs are synchronized and overlaid.

    Args:
        data_frame (pd.DataFrame): The data source for the chart.
        x (str): The column name to be used for the x-axis.
        y_column (str): The column name to be used for the y-axis 1, representing the column chart.
        y_line (str): The column name to be used for the y-axis 2, representing the line chart.

    Returns:
        go.Figure: : A Plotly Figure object representing the combined column and line chart.

    """
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    fig.add_trace(
        go.Bar(x=data_frame[x], y=data_frame[y_column], name=y_column),
        secondary_y=False,
    )

    fig.add_trace(
        go.Scatter(x=data_frame[x], y=data_frame[y_line], name=y_line),
        secondary_y=True,
    )

    fig.update_layout(
        xaxis={"type": "category", "title": x},
        yaxis={"tickmode": "sync", "title": y_column},
        yaxis2={"tickmode": "sync", "overlaying": "y", "title": y_line},
    )

    return fig


@capture("graph")
def categorical_column(data_frame: pd.DataFrame, x: str, y: str):
    """Creates a column chart where the x-axis values are converted to category type."""
    fig = px.bar(
        data_frame,
        x=x,
        y=y,
    )
    # So ticks are aligned with bars when xaxes values are numbers (e.g. years)
    fig.update_xaxes(type="category")
    return fig



"""Magnitude charts."""

import vizro.models as vm
import vizro.plotly.express as px

from pages._pages_utils import PAGE_GRID, gapminder, iris, make_code_clipboard_from_py_file, tips

bar = vm.Page(
    title="Bar",
    path="magnitude/bar",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a bar chart?

            A bar chart displays bars with lengths proportional to the values they represent. One axis shows the
            categories to compare, and the other provides a value scale starting from zero.

            &nbsp;

            #### When should I use it?

            Use a bar chart to help your audience compare sizes and identify patterns in categorical data, such as
            **how many?** in each category. Arrange the bars in any order to fit the message you want to emphasize.
            Ensure clear labeling, especially with many bars, and consider using a legend or abbreviations with fuller
            descriptions below.
        """
        ),
        vm.Graph(
            figure=px.bar(
                gapminder.query(
                    "year == 2007 and country.isin(['United States', 'Pakistan', 'India', 'China', 'Indonesia'])"
                ),
                x="pop",
                y="country",
                orientation="h",
            )
        ),
        make_code_clipboard_from_py_file("bar.py"),
    ],
)

# Note: Code example for magnitude/column differs from time/column. The text description is the same.
column = vm.Page(
    id="magnitude-column",
    path="magnitude/column",
    title="Column",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
                #### What is a column chart?

                A column chart is a type of bar chart where data is represented with vertical columns. Each
                column's height corresponds to the value it represents, with the y-axis starting from zero.

                &nbsp;

                #### When should I use it?

                Use a column chart to compare sizes and identify patterns in categorical data, including time-based
                data. Arrange columns to fit your message, and for time-based data, order them chronologically to
                highlight trends. Ensure clear labeling, especially with many categories, and consider using a legend
                or abbreviations with fuller descriptions below.
        """
        ),
        vm.Graph(
            figure=px.bar(
                gapminder.query(
                    "year == 2007 and country.isin(['United States', 'Pakistan', 'India', 'China', 'Indonesia'])"
                ),
                y="pop",
                x="country",
            )
        ),
        make_code_clipboard_from_py_file("magnitude_column.py"),
    ],
)

paired_bar = vm.Page(
    title="Paired bar",
    path="magnitude/paired-bar",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a paired bar chart?

            A paired bar chart, also known as a grouped bar chart, displays bars grouped together in pairs for each
            category, with the lengths of the bars proportional to the values they represent. One axis shows the
            categories to compare, while the other provides a value scale starting from zero.

            &nbsp;

            #### When should I use it?

            Use a paired bar chart to compare multiple sets of data within the same categories. This type of chart is
            particularly useful for highlighting differences and similarities between groups, such as **how many?** in
            each category across different groups. Arrange the paired bars clearly to fit the message you want to
            emphasize. Ensure clear labeling, especially with many bars, and consider using a legend or abbreviations
            with fuller descriptions below.
        """
        ),
        vm.Graph(
            figure=px.histogram(
                tips,
                y="day",
                x="total_bill",
                color="sex",
                barmode="group",
                orientation="h",
                category_orders={"day": ["Thur", "Fri", "Sat", "Sun"]},
            ),
        ),
        make_code_clipboard_from_py_file("paired_bar.py"),
    ],
)

paired_column = vm.Page(
    title="Paired column",
    path="magnitude/paired-column",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a paired column chart?

            A paired column chart, also known as a grouped column chart, displays columns grouped together in pairs for
            each category, with the heights of the columns proportional to the values they represent. One axis shows the
            categories to compare, while the other provides a value scale starting from zero.

            &nbsp;

            #### When should I use it?

            Use a paired column chart to compare multiple sets of data within the same categories. This type of chart is
            particularly useful for highlighting differences and similarities between groups, such as **how many?** in
            each category across different groups. Arrange the paired columns clearly to fit the message you want to
            emphasize. Ensure clear labeling, especially with many columns, and consider using a legend or abbreviations
            with fuller descriptions below.
        """
        ),
        vm.Graph(
            figure=px.histogram(
                tips,
                x="day",
                y="total_bill",
                color="sex",
                barmode="group",
                category_orders={"day": ["Thur", "Fri", "Sat", "Sun"]},
            ),
        ),
        make_code_clipboard_from_py_file("paired_column.py"),
    ],
)

parallel_coordinates = vm.Page(
    path="magnitude/parallel-coordinates ",
    title="Parallel coordinates",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
                #### What is a parallel coordinates chart?

                A parallel coordinates chart is a type of data visualization used to plot multivariate numerical data.
                Each axis represents a different variable, and lines connecting the axes indicate the values of
                individual data points across these variables.

                &nbsp;

                #### When should I use it?

                Use a parallel coordinates chart to explore relationships and patterns in high-dimensional data.
                This chart is particularly useful for comparing multiple variables simultaneously and identifying
                correlations or clusters within the data. Ensure clear labeling of each axis and consider using color
                coding to distinguish between different data points or groups.
        """
        ),
        vm.Graph(
            figure=px.parallel_coordinates(
                iris, color="species_id", dimensions=["sepal_width", "sepal_length", "petal_width", "petal_length"]
            )
        ),
        make_code_clipboard_from_py_file("parallel_coordinates.py"),
    ],
)

pages = [bar, column, paired_bar, paired_column, parallel_coordinates]



# ruff: noqa: F403, F405
import os
import runpy
from pathlib import Path

import chromedriver_autoinstaller
import pytest
from vizro import Vizro


# Taken from https://github.com/pytest-dev/pytest/issues/363#issuecomment-1335631998.
@pytest.fixture(scope="module")
def monkeypatch_session():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="module", autouse=True)
def setup_integration_test_environment(monkeypatch_session):
    # Dash debug mode seems to interfere with the tests, so we disable it here. Note "false" as a string is correct.
    monkeypatch_session.setenv("DASH_DEBUG", "false")
    # We only need to install chromedriver outside CI.
    if not os.getenv("CI"):
        chromedriver_autoinstaller.install()


@pytest.fixture
def dashboard(request, monkeypatch):
    example_directory = request.getfixturevalue("example_path") / request.getfixturevalue("version")
    monkeypatch.chdir(example_directory)
    monkeypatch.syspath_prepend(example_directory)
    return runpy.run_path("app.py")["dashboard"]
    # Both run_path and run_module contaminate sys.modules, so we need to undo this in order to avoid interference
    # between tests. However, if you do this then importlib.import_module seems to cause the problem due to mysterious
    # reasons. The current system should work well so long as there's no sub-packages with clashing names in the
    # examples.


examples_path = Path(__file__).parents[2] / "examples"


# Ignore deprecation warning until this is solved: https://github.com/plotly/dash/issues/2590
# The `features` examples do add_type, which ideally we would clean up afterwards to restore vizro.models to its
# previous state. Since we don't currently do this, `hatch run test` fails.
# This is difficult to fix fully by un-importing vizro.models though, since we use `import vizro.models as vm` - see
# https://stackoverflow.com/questions/437589/how-do-i-unload-reload-a-python-module.
@pytest.mark.filterwarnings("ignore:HTTPResponse.getheader()")
# Ignore as it doesn't affect the test run
@pytest.mark.filterwarnings("ignore::pytest.PytestUnhandledThreadExceptionWarning")
@pytest.mark.parametrize(
    "example_path, version",
    [
        # KPI example is not included as it will be moved to HuggingFace over time.
        # Visual vocabulary is not included since it means installing black in the testing environment.
        # It will move to HuggingFace in due course anyway.
        (examples_path / "scratch_dev", ""),
        (examples_path / "scratch_dev", "yaml_version"),
        (examples_path / "dev", ""),
        (examples_path / "dev", "yaml_version"),
    ],
    ids=str,
)
def test_dashboard(dash_duo, example_path, dashboard, version):
    app = Vizro(assets_folder=example_path / "assets").build(dashboard).dash
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



"""File to simulate imports from other modules."""

import vizro.models as vm
import vizro.plotly.express as px
from dash_ag_grid import AgGrid
from vizro.actions import export_data
from vizro.models.types import capture
from vizro.tables import dash_ag_grid

df = px.data.iris()


@capture("ag_grid")
def my_custom_aggrid(chosen_columns, data_frame=None):
    """Custom ag_grid."""
    defaults = {
        "className": "ag-theme-quartz-dark ag-theme-vizro",
        "defaultColDef": {
            "resizable": True,
            "sortable": True,
            "filter": True,
            "filterParams": {
                "buttons": ["apply", "reset"],
                "closeOnApply": True,
            },
            "flex": 1,
            "minWidth": 70,
        },
        "style": {"height": "100%"},
    }
    return AgGrid(
        columnDefs=[{"field": col} for col in chosen_columns], rowData=data_frame.to_dict("records"), **defaults
    )


page2 = vm.Page(
    title="Page2",
    components=[
        vm.Graph(id="hist_chart2", figure=px.histogram(df, x="sepal_width", color="species")),
        vm.AgGrid(figure=my_custom_aggrid(data_frame="iris", chosen_columns=["sepal_width", "sepal_length"])),
        vm.AgGrid(figure=dash_ag_grid(data_frame="iris")),
        vm.Button(
            text="Export data",
            actions=[
                vm.Action(function=export_data()),
                vm.Action(
                    function=export_data(
                        file_format="xlsx",
                    )
                ),
            ],
        ),
    ],
)



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

tips = px.data.tips()

page = vm.Page(
    title="Boxplot",
    components=[
        vm.Graph(figure=px.boxplot(tips, y="total_bill", x="day", color="day")),
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



from contextlib import contextmanager
from typing import Any, List, Mapping, Optional, Set, Type, Union

try:
    from pydantic.v1 import BaseModel, Field, validator
    from pydantic.v1.fields import SHAPE_LIST, ModelField
    from pydantic.v1.typing import get_args
except ImportError:  # pragma: no cov
    from pydantic import BaseModel, Field, validator
    from pydantic.fields import SHAPE_LIST, ModelField
    from pydantic.typing import get_args


import inspect
import logging
import subprocess
import textwrap

from typing_extensions import Annotated

from vizro.managers import model_manager
from vizro.models._models_utils import REPLACEMENT_STRINGS, _log_call

ACTIONS_CHAIN = "ActionsChain"
ACTION = "actions"

TO_PYTHON_TEMPLATE = """
############ Imports ##############
import vizro.plotly.express as px
import vizro.tables as vt
import vizro.models as vm
import vizro.actions as va
from vizro import Vizro
from vizro.models.types import capture
from vizro.managers import data_manager
{extra_imports}

{callable_defs_template}
{data_settings_template}

########### Model code ############
{code}
"""

CALLABLE_TEMPLATE = """
####### Function definitions ######
{callable_defs}
"""

DATA_TEMPLATE = """
####### Data Manager Settings #####
#######!!! UNCOMMENT BELOW !!!#####
# from vizro.managers import data_manager
{data_setting}
"""

# Global variable to dictate whether VizroBaseModel.dict should be patched to work for _to_python.
# This is always False outside the _patch_vizro_base_model_dict context manager to ensure that, unless explicitly
# called for, dict behavior is unmodified from pydantic's default.
_PATCH_VIZRO_BASE_MODEL_DICT = False


@contextmanager
def _patch_vizro_base_model_dict():
    global _PATCH_VIZRO_BASE_MODEL_DICT  # noqa
    _PATCH_VIZRO_BASE_MODEL_DICT = True
    try:
        yield
    finally:
        _PATCH_VIZRO_BASE_MODEL_DICT = False


def _format_and_lint(code_string: str) -> str:
    # Tracking https://github.com/astral-sh/ruff/issues/659 for proper python API
    # Good example: https://github.com/astral-sh/ruff/issues/8401#issuecomment-1788806462
    linted = subprocess.check_output(
        ["ruff", "check", "--fix", "--exit-zero", "--silent", "--isolated", "-"], input=code_string, encoding="utf-8"
    )
    formatted = subprocess.check_output(
        ["ruff", "format", "--silent", "--isolated", "-"], input=linted, encoding="utf-8"
    )
    return formatted


def _dict_to_python(object: Any) -> str:
    from vizro.models.types import CapturedCallable

    if isinstance(object, dict) and "__vizro_model__" in object:
        __vizro_model__ = object.pop("__vizro_model__")

        # This is required to back-engineer the actions chains. It is easier to handle in the string conversion here
        # than in the dict creation, because we end up with nested lists when being forced to return a list of actions.
        # If we handle it in dict of vm.BaseModel, then we created an unexpected dict return.
        if __vizro_model__ == ACTIONS_CHAIN:
            action_data = object[ACTION]
            return ", ".join(_dict_to_python(item) for item in action_data)
        else:
            # This is very similar to doing repr but includes the vm. prefix and calls _object_to_python_code
            # rather than repr recursively.
            fields = ", ".join(f"{field_name}={_dict_to_python(value)}" for field_name, value in object.items())
            return f"vm.{__vizro_model__}({fields})"
    elif isinstance(object, dict):
        fields = ", ".join(f"{field_name}={_dict_to_python(value)}" for field_name, value in object.items())
        return "{" + fields + "}"
    elif isinstance(object, list):
        # Need to do this manually to avoid extra quotation marks that arise when doing repr(List).
        code_string = ", ".join(_dict_to_python(item) for item in object)
        return f"[{code_string}]"
    elif isinstance(object, CapturedCallable):
        return object.__repr_clean__()
    else:
        return repr(object)


# The two extract helper functions may not work when we refactor the model_manager to work differently when models
# are created. An alternative approach to iterating through the model_manager is to recurse through the object as
# is done in the _dict_to_python function.
# Note also that these functions find also unintended model_manager additions, a known but accepted limitation.
def _extract_captured_callable_source() -> Set[str]:
    from vizro.models.types import CapturedCallable

    captured_callable_sources = set()
    for model_id in model_manager:
        for _, value in model_manager[model_id]:
            if isinstance(value, CapturedCallable) and not any(
                # Check to see if the captured callable does use a cleaned module string, if yes then
                # we can assume that the source code can be imported via Vizro, and thus does not need to be defined
                value.__repr_clean__().startswith(new)
                for _, new in REPLACEMENT_STRINGS.items()
            ):
                try:
                    source = textwrap.dedent(inspect.getsource(value._function))
                    captured_callable_sources.add(source)
                except OSError:
                    # OSError is raised when the source code is not available. This is expected
                    # for built-in functions or dynamically defined functions (via exec or eval).
                    logging.warning(f"Could not extract source for {value._function}. Definition will not be included.")
                    pass
    return captured_callable_sources


def _extract_captured_callable_data_info() -> Set[str]:
    from vizro.models.types import CapturedCallable

    return {
        f'# data_manager["{value["data_frame"]}"] = ===> Fill in here <==='
        for model_id in model_manager
        for _, value in model_manager[model_id]
        if isinstance(value, CapturedCallable)
        if "data_frame" in value._arguments
    }


class VizroBaseModel(BaseModel):
    """All models that are registered to the model manager should inherit from this class.

    Args:
        id (str): ID to identify model. Must be unique throughout the whole dashboard. Defaults to `""`.
            When no ID is chosen, ID will be automatically generated.

    """

    id: str = Field(
        "",
        description="ID to identify model. Must be unique throughout the whole dashboard."
        "When no ID is chosen, ID will be automatically generated.",
    )

    @validator("id", always=True)
    def set_id(cls, id) -> str:
        return id or model_manager._generate_id()

    @_log_call
    def __init__(self, **data: Any):
        """Adds this model instance to the model manager."""
        # Note this runs after the set_id validator, so self.id is available here. In pydantic v2 we should do this
        # using the new model_post_init method to avoid overriding __init__.
        super().__init__(**data)
        model_manager[self.id] = self

    @classmethod
    def add_type(cls, field_name: str, new_type: Type[Any]):
        """Adds a new type to an existing field based on a discriminated union.

        Args:
            field_name: Field that new type will be added to
            new_type: New type to add to discriminated union

        """

        def _add_to_discriminated_union(union):
            # Returning Annotated here isn't strictly necessary but feels safer because the new discriminated union
            # will then be annotated the same way as the old one.
            args = get_args(union)
            # args[0] is the original union, e.g. Union[Filter, Parameter]. args[1] is the FieldInfo annotated metadata.
            return Annotated[Union[args[0], new_type], args[1]]

        def _is_discriminated_union(field):
            # Really this should be done as follows:
            # return field.discriminator_key is not None
            # However, this does not work with Optional[DiscriminatedUnion]. See also TestOptionalDiscriminatedUnion.
            return hasattr(field.outer_type_, "__metadata__") and get_args(field.outer_type_)[1].discriminator

        field = cls.__fields__[field_name]
        sub_field = field.sub_fields[0] if field.shape == SHAPE_LIST else None

        if _is_discriminated_union(field):
            # Field itself is a non-optional discriminated union, e.g. selector: SelectorType or Optional[SelectorType].
            new_annotation = _add_to_discriminated_union(field.outer_type_)
        elif sub_field is not None and _is_discriminated_union(sub_field):
            # Field is a list of discriminated union e.g. components: List[ComponentType].
            new_annotation = List[_add_to_discriminated_union(sub_field.outer_type_)]  # type: ignore[misc]
        else:
            raise ValueError(
                f"Field '{field_name}' must be a discriminated union or list of discriminated union type. "
                "You probably do not need to call add_type to use your new type."
            )

        cls.__fields__[field_name] = ModelField(
            name=field.name,
            type_=new_annotation,
            class_validators=field.class_validators,
            model_config=field.model_config,
            default=field.default,
            default_factory=field.default_factory,
            required=field.required,
            final=field.final,
            alias=field.alias,
            field_info=field.field_info,
        )

        # We need to resolve all ForwardRefs again e.g. in the case of Page, which requires update_forward_refs in
        # vizro.models. The vm.__dict__.copy() is inspired by pydantic's own implementation of update_forward_refs and
        # effectively replaces all ForwardRefs defined in vizro.models.
        import vizro.models as vm

        cls.update_forward_refs(**vm.__dict__.copy())
        new_type.update_forward_refs(**vm.__dict__.copy())

    def dict(self, **kwargs):
        global _PATCH_VIZRO_BASE_MODEL_DICT  # noqa
        if not _PATCH_VIZRO_BASE_MODEL_DICT:
            # Whenever dict is called outside _patch_vizro_base_model_dict, we don't modify the behavior of the dict.
            return super().dict(**kwargs)

        # When used in _to_python, we overwrite pydantic's own `dict` method to add __vizro_model__ to the dictionary
        # and to exclude fields specified dynamically in __vizro_exclude_fields__.
        # To get exclude as an argument is a bit fiddly because this function is called recursively inside pydantic,
        # which sets exclude=None by default.
        if kwargs.get("exclude") is None:
            kwargs["exclude"] = self.__vizro_exclude_fields__()
        _dict = super().dict(**kwargs)
        _dict["__vizro_model__"] = self.__class__.__name__
        return _dict

    # This is like pydantic's own __exclude_fields__ but safer to use (it looks like __exclude_fields__ no longer
    # exists in pydantic v2).
    # Root validators with pre=True are always included, even when exclude_default=True, and so this is needed
    # to potentially exclude fields set this way, like Page.id.
    def __vizro_exclude_fields__(self) -> Optional[Union[Set[str], Mapping[str, Any]]]:
        return None

    def _to_python(
        self, extra_imports: Optional[Set[str]] = None, extra_callable_defs: Optional[Set[str]] = None
    ) -> str:
        """Converts a Vizro model to the Python code that would create it.

        Args:
            extra_imports: Extra imports to add to the Python code. Provide as a set of complete import strings.
            extra_callable_defs: Extra callable definitions to add to the Python code. Provide as a set of complete
                function definitions.

        Returns:
            str: Python code to create the Vizro model.

        Examples:
            Simple usage example with card model.

            >>> import vizro.models as vm
            >>> card = vm.Card(text="Hello, world!")
            >>> print(card._to_python())

            Further options include adding extra imports and callable definitions. These will be included in the
            returned python string.

            >>> print(card._to_python(
            ...    extra_imports={"from typing import List"},
            ...    extra_callable_defs={"def test(foo:List[str]): return foo"}))

        """
        # Imports
        extra_imports_concat = "\n".join(extra_imports) if extra_imports else None

        # CapturedCallable definitions - NOTE that order is not guaranteed
        callable_defs_set = _extract_captured_callable_source() | (
            extra_callable_defs if extra_callable_defs else set()
        )
        callable_defs_concat = "\n".join(callable_defs_set) if callable_defs_set else None

        # Data Manager
        data_defs_set = _extract_captured_callable_data_info()
        data_defs_concat = "\n".join(data_defs_set) if data_defs_set else None

        # Model code
        with _patch_vizro_base_model_dict():
            model_dict = self.dict(exclude_unset=True)

        model_code = "model = " + _dict_to_python(model_dict)

        # Concatenate and lint code
        callable_defs_template = (
            CALLABLE_TEMPLATE.format(callable_defs=callable_defs_concat) if callable_defs_concat else ""
        )
        data_settings_template = DATA_TEMPLATE.format(data_setting=data_defs_concat) if data_defs_concat else ""
        unformatted_code = TO_PYTHON_TEMPLATE.format(
            code=model_code,
            extra_imports=extra_imports_concat if extra_imports_concat else "",
            callable_defs_template=callable_defs_template,
            data_settings_template=data_settings_template,
        )
        try:
            return _format_and_lint(unformatted_code)
        except Exception:
            logging.exception("Code formatting failed; returning unformatted code")
            return unformatted_code

    class Config:
        extra = "forbid"  # Good for spotting user typos and being strict.
        smart_union = True  # Makes unions work without unexpected coercion (will be the default in pydantic v2).
        validate_assignment = True  # Run validators when a field is assigned after model instantiation.
        copy_on_model_validation = "none"  # Don't copy sub-models. Essential for the model_manager to work correctly.



from __future__ import annotations

import logging
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, List

import dash
import flask
import plotly.io as pio
from flask_caching import SimpleCache

from vizro._constants import STATIC_URL_PREFIX
from vizro.managers import data_manager, model_manager
from vizro.models import Dashboard

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    # These are built into wsgiref.types for Python 3.11 onwards.
    from _typeshed.wsgi import StartResponse, WSGIEnvironment


class Vizro:
    """The main class of the `vizro` package."""

    def __init__(self, **kwargs):
        """Initializes Dash app, stored in `self.dash`.

        Args:
            **kwargs : Passed through to `Dash.__init__`, e.g. `assets_folder`, `url_base_pathname`. See
                [Dash documentation](https://dash.plotly.com/reference#dash.dash) for possible arguments.

        """
        # Setting suppress_callback_exceptions=True for the following reasons:
        # 1. Prevents the following Dash exception when using html.Div as placeholders in build methods:
        #    "Property 'cellClicked' was used with component ID '__input_ag_grid_id' in one of the Input
        #    items of a callback. This ID is assigned to a dash_html_components.Div component in the layout,
        #    which does not support this property."
        # 2. Improves performance by bypassing layout validation.
        self.dash = dash.Dash(
            **kwargs,
            pages_folder="",
            suppress_callback_exceptions=True,
            title="Vizro",
            use_pages=True,
        )

        # Include Vizro assets (in the static folder) as external scripts and stylesheets. We extend self.dash.config
        # objects so the user can specify additional external_scripts and external_stylesheets via kwargs.
        vizro_assets_folder = Path(__file__).with_name("static")
        requests_pathname_prefix = self.dash.config.requests_pathname_prefix
        # Exclude vizro/css/figures.css since these are distributed through the vizro._css_dist mechanism.
        # In future we will probably handle all assets this way and none of this code will be required.
        vizro_css = [
            requests_pathname_prefix + path
            for path in self._get_external_assets(vizro_assets_folder, "css")
            if path != "vizro/css/figures.css"
        ]

        # Ensure vizro-bootstrap.min.css is loaded in first to allow overwrites
        vizro_css.sort(key=lambda x: not x.endswith("vizro-bootstrap.min.css"))

        vizro_js = [
            {"src": requests_pathname_prefix + path, "type": "module"}
            for path in self._get_external_assets(vizro_assets_folder, "js")
        ]
        self.dash.config.external_stylesheets.extend(vizro_css)
        self.dash.config.external_scripts.extend(vizro_js)

        # Serve all assets (including files other than css and js) that live in vizro_assets_folder at the
        # route /vizro. Based on code in Dash.init_app that serves assets_folder. This respects the case that the
        # dashboard is not hosted at the root of the server, e.g. http://www.example.com/dashboard/vizro.
        routes_pathname_prefix = self.dash.config.routes_pathname_prefix
        blueprint_prefix = routes_pathname_prefix.replace("/", "_").replace(".", "_")
        self.dash.server.register_blueprint(
            flask.Blueprint(
                f"{blueprint_prefix}vizro_assets",
                self.dash.config.name,
                static_folder=vizro_assets_folder,
                static_url_path=routes_pathname_prefix + STATIC_URL_PREFIX,
            )
        )

        data_manager.cache.init_app(self.dash.server)

    def build(self, dashboard: Dashboard):
        """Builds the `dashboard`.

        Args:
            dashboard (Dashboard): [`Dashboard`][vizro.models.Dashboard] object.

        Returns:
            self: Vizro app

        """
        # Note Dash.index uses self.dash.title instead of self.dash.app.config.title.
        if dashboard.title:
            self.dash.title = dashboard.title

        # Set global template to vizro_light or vizro_dark.
        # The choice between these is generally meaningless because chart colors in the two are identical, and
        # everything else gets overridden in the post-fig creation layout.template update in Graph.__call__ and the
        # clientside theme selector callback.
        # Note this setting of global template isn't undone anywhere. If we really wanted to then we could try and
        # put in some teardown code, but it would probably never be 100% reliable. Vizro._reset can't do this well
        # either because it's a staticmethod so can't access self.old_theme (though we could use a global variable to
        # store it). Remember this template setting can't go in run() though since it's needed even in deployment.
        # Probably the best solution if we do want to fix this would be to have two separate paths that are followed:
        # 1. In deployment (or just outside Jupyter?), set the theme here and never revert it.
        # 2. In other contexts, use context manager in run method.
        pio.templates.default = dashboard.theme

        # Note that model instantiation and pre_build are independent of Dash.
        self._pre_build()

        self.dash.layout = dashboard.build()

        return self

    def run(self, *args, **kwargs):  # if type annotated, mkdocstring stops seeing the class
        """Runs the dashboard.

        Args:
            *args : Passed through to `dash.run`.
            **kwargs : Passed through to `dash.run`.

        """
        data_manager._frozen_state = True
        model_manager._frozen_state = True

        if kwargs.get("processes", 1) > 1 and type(data_manager.cache.cache) is SimpleCache:
            warnings.warn(
                "`SimpleCache` is designed to support only single process environments. If you would like to use "
                "multiple processes then you should change to a cache that supports it such as `FileSystemCache` or "
                "`RedisCache`."
            )

        self.dash.run(*args, **kwargs)

    @staticmethod
    def _pre_build():
        """Runs pre_build method on all models in the model_manager."""
        # Note that a pre_build method can itself add a model (e.g. an Action) to the model manager, and so we need to
        # iterate through set(model_manager) rather than model_manager itself or we loop through something that
        # changes size.
        # Any models that are created during the pre-build process *will not* themselves have pre_build run on them.
        # In future may add a second pre_build loop after the first one.
        for model_id in set(model_manager):
            model = model_manager[model_id]
            if hasattr(model, "pre_build"):
                model.pre_build()

    def __call__(self, environ: WSGIEnvironment, start_response: StartResponse) -> Iterable[bytes]:
        """Implements WSGI application interface.

        This means you can do e.g. gunicorn app:app without needing to manually define server = app.dash.server.
        """
        return self.dash.server(environ, start_response)

    @staticmethod
    def _reset():
        """Private method that clears all state in the `Vizro` app.

        This deliberately does not clear the data manager cache - see comments in data_manager._clear for
        explanation.
        """
        data_manager._clear()
        model_manager._clear()
        dash._callback.GLOBAL_CALLBACK_LIST = []
        dash._callback.GLOBAL_CALLBACK_MAP = {}
        dash._callback.GLOBAL_INLINE_SCRIPTS = []
        dash.page_registry.clear()
        dash._pages.CONFIG.clear()
        dash._pages.CONFIG.__dict__.clear()

    @staticmethod
    def _get_external_assets(folder: Path, extension: str) -> List[str]:
        """Returns a list of paths to assets with given `extension` in `folder`, prefixed with `STATIC_URL_PREFIX`.

        e.g. with STATIC_URL_PREFIX="vizro", extension="css", folder="/path/to/vizro/vizro-core/src/vizro/static",
        we will get ["vizro/css/accordion.css", "vizro/css/button.css", ...].
        """
        return sorted(
            (STATIC_URL_PREFIX / path.relative_to(folder)).as_posix() for path in folder.rglob(f"*.{extension}")
        )



"""Spatial charts."""

import vizro.models as vm
import vizro.plotly.express as px

from pages._pages_utils import PAGE_GRID, gapminder, make_code_clipboard_from_py_file

choropleth = vm.Page(
    title="Choropleth",
    path="spatial/choropleth",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
            #### What is a choropleth map?

            A choropleth map is a map in which geographical areas are colored, shaded or patterned in relation to a
            specific data variable.

            &nbsp;

            #### When should I use it?

            Use a chloropleth map when you wish to show how a measurement varies across a geographic area, or to show
            variability or patterns within a region. Typically, you will blend one color into another, take a color
            shade from light to dark, or introduce patterns to depict the variation in the data.

            Be aware that it may be difficult for your audience to accurately read or compare values on the map
            depicted by color.

        """
        ),
        vm.Graph(
            figure=px.choropleth(
                gapminder.query("year == 2007"),
                locations="iso_alpha",
                color="lifeExp",
                hover_name="country",
            )
        ),
        make_code_clipboard_from_py_file("choropleth.py"),
    ],
)

pages = [choropleth]



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

iris = px.data.iris()

page = vm.Page(
    title="Parallel coordinates",
    components=[
        vm.Graph(
            figure=px.parallel_coordinates(
                iris, color="species_id", dimensions=["sepal_width", "sepal_length", "petal_width", "petal_length"]
            )
        )
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import pandas as pd
import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro
from vizro.models.types import capture

gapminder = px.data.gapminder()


@capture("graph")
def categorical_column(data_frame: pd.DataFrame, x: str, y: str):
    fig = px.bar(
        data_frame,
        x=x,
        y=y,
    )
    # So ticks are aligned with bars when xaxes values are numbers (e.g. years)
    fig.update_xaxes(type="category")
    return fig


page = vm.Page(
    title="Column",
    components=[
        vm.Graph(
            figure=categorical_column(
                gapminder.query("country == 'Nigeria' and year > 1970"),
                y="lifeExp",
                x="year",
            )
        )
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

iris = px.data.iris()

page = vm.Page(
    title="Scatter",
    components=[vm.Graph(figure=px.scatter(iris, x="sepal_width", y="sepal_length", color="species"))],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""The data manager handles access to all DataFrames used in a Vizro app."""

from __future__ import annotations

import functools
import logging
import os
import warnings
from functools import partial
from typing import Callable, Dict, Optional, Union

import pandas as pd
import wrapt
from flask_caching import Cache

from vizro.managers._managers_utils import _state_modifier

logger = logging.getLogger(__name__)


# Really DataSourceName should be NewType and not just aliases but then for a user's code to type check
# correctly they would need to cast all strings to these types.
DataSourceName = str
pd_DataFrameCallable = Callable[..., pd.DataFrame]


# TODO: consider merging with model_utils _log_call. Using wrapt.decorator is probably better than functools here.
#  Might need messages that run before/after the wrapped function call.
# Follows the pattern recommended in https://wrapt.readthedocs.io/en/latest/decorators.html#decorators-with-arguments
# for making a wrapt.decorator with arguments.
def _log_call(message: str):
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        logger.debug(message)
        # Might be useful if merged with model_utils _log_call:
        # logging.debug(message.format(wrapped=wrapped, instance=instance, args=args, kwargs=kwargs))
        return wrapped(*args, **kwargs)

    return wrapper


class _DynamicData:
    """Wrapper for a pd_DataFrameCallable, i.e. a function that produces a pandas DataFrame.

    Crucially this means that the data can be refreshed during runtime, since the loading function can be re-run.

    This is currently private since it's not expected that a user would instantiate it directly. Instead, you would use
    through the data_manager interface as follows:
        >>> def dynamic_data():
        >>>     return pd.read_csv("dynamic_data.csv")
        >>> data_manager["dynamic_data"] = dynamic_data
        >>> data_manager["dynamic_data"].timeout = 5  # if you want to change the cache timeout to 5 seconds

    Possibly in future, this will become a public class so you could directly do:
        >>> data_manager["dynamic_data"] = DynamicData(dynamic_data, timeout=5)

    At this point we might like to disable the behavior so that data_manager setitem and getitem handle the same
    object rather than doing an implicit conversion to _DynamicData.
    """

    def __init__(self, load_data: pd_DataFrameCallable):
        self.__load_data: pd_DataFrameCallable = load_data
        self.timeout: Optional[int] = None
        # We might also want a self.cache_arguments dictionary in future that allows user to customize more than just
        # timeout, but no rush to do this since other arguments are unlikely to be useful.

    def load(self, *args, **kwargs) -> pd.DataFrame:
        """Loads data."""
        # Data source name can be extracted from the function's name since it was added there in DataManager.__setitem__
        logger.debug(
            "Looking in cache for data source %s on process %s",
            self.__load_data.__name__.rpartition(".")[-1],
            os.getpid(),
        )
        # We don't memoize the load method itself as this is tricky to get working fully when load is called with
        # arguments, since we need the signature of the memoized function to match that of load_data. See
        # https://github.com/GrahamDumpleton/wrapt/issues/263.
        # It's also difficult to get memoize working correctly with bound methods anyway - see comment in
        # DataManager.__setitem__. It's much easier to ensure that self.__load_data is always just a function.
        if data_manager._cache_has_app:
            # This includes the case of NullCache.
            load_data = _log_call("Cache miss; reloading data")(self.__load_data)
            load_data = data_manager.cache.memoize(timeout=self.timeout)(load_data)
        else:
            logger.debug("Cache not active; reloading data")
            load_data = self.__load_data
        return load_data(*args, **kwargs)


class _StaticData:
    """Wrapper for a pd.DataFrame. This data cannot be updated during runtime.

    This is currently private since it's not expected that a user would instantiate it directly. Instead, you would use
    through the data_manager interface as follows:
         >>> data_manager["static_data"] = pd.read_csv("static_data.csv")

    This class does not have much functionality but exists for a couple of reasons:
        1. to align interface with _DynamicData by providing a load method so that fetching data from data_manager can
        transparently handle loading both _StaticData and _DynamicData without needing switching logic
        2. to raise a clear error message if a user tries to set a timeout on the data source
    """

    def __init__(self, data: pd.DataFrame):
        self.__data = data

    def load(self) -> pd.DataFrame:
        """Loads data.

        Returns a copy of the data. This is not necessary if we are careful to not do any inplace=True operations,
        but safest to leave it here, e.g. in case a user-defined action mutates the data. To be even safer we could
        additionally (but not instead) copy data when setting it in __init__ but this consumes more memory and is not
        necessary so long as data is only ever accessed through the intended API of data_manager["static_data"].load().
        """
        return self.__data.copy()

    def __setattr__(self, name, value):
        # Any attributes that are only relevant for _DynamicData should go here to raise a clear error message.
        if name in ["timeout"]:
            raise AttributeError(
                f"Static data that is a pandas.DataFrame itself does not support {name}; you should instead use a "
                "dynamic data source that is a function that returns a pandas.DataFrame."
            )
        super().__setattr__(name, value)


class DataManager:
    """Object to handle all data for the `vizro` application.

    Examples
        >>> # Static data that cannot be refreshed during runtime
        >>> data_manager["data"] = pd.read_csv("data.csv")
        >>> # Data that can be refreshed during runtime
        >>> def dynamic_data():
        >>>     return pd.read_csv("dynamic_data.csv")
        >>> data_manager["dynamic_data"] = dynamic_data
        >>> data_manager["dynamic_data"].timeout = 5  # if you want to change the cache timeout to 5 seconds

    """

    def __init__(self):
        self.__data: Dict[DataSourceName, Union[_DynamicData, _StaticData]] = {}
        self._frozen_state = False
        self.cache = Cache(config={"CACHE_TYPE": "NullCache"})
        # In future, possibly we will accept just a config dict. Would need to work out whether to handle merging with
        # default values though. We would do this with something like this:
        # def __set_cache(self, cache_config):
        #     if not isinstance(cache, Cache):
        #         self._cache = Cache(**cache)
        #     self._cache = value
        # _cache = property(fset=__set_cache)

    @_state_modifier
    def __setitem__(self, name: DataSourceName, data: Union[pd.DataFrame, pd_DataFrameCallable]):
        """Adds `data` to the `DataManager` with key `name`."""
        if callable(data):
            # __qualname__ is required by flask-caching (even if we specify our own make_name) but
            # not defined for partial functions and just '<lambda>' for lambda functions. Defining __qualname__
            # means it's possible to have non-interfering caches for lambda functions (similarly if we
            # end up using CapturedCallable, or that could instead set its own __qualname__).
            # We handle __name__ the same way even though it's not currently essential to functioning of flask-caching
            # in case they change the underlying implementation to use it.
            # We use partial to effectively make an independent copy of the underlying data function. This means that
            # it's possible to set __qualname__ independently for each data source. This is not essential for
            # functions other than lambda, but it is essential for bound methods, as flask-caching cannot easily
            # independently timeout different instances with different bound methods but the same underlying function
            # data.__func__. If we don't do this then the bound method case needs some uglier hacking to make work
            # correctly - see https://github.com/mckinsey/vizro/blob/abb7eebb230ba7e6cfdf6150dc56b211a78b1cd5/
            # vizro-core/src/vizro/managers/_data_manager.py.
            # Once partial has been used, all dynamic data sources are on equal footing since they're all treated as
            # functions rather than bound methods, e.g. by flask_caching.utils.function_namespace. This makes it much
            # simpler to use flask-caching reliably.
            # It's important the __qualname__ is the same across all workers, so use the data source name rather than
            # e.g. the repr method that includes the id of the instance so would only work in the case that gunicorn is
            # running with --preload.
            # __module__ is also required in flask_caching.utils.function_namespace and not defined for partial
            # functions in some versions of Python.
            # update_wrapper ensures that __module__, __name__, __qualname__, __annotations__ and __doc__ are
            # assigned to the new partial(data) the same as they were in data. This isn't strictly necessary but makes
            # inspecting these functions easier.
            data = functools.update_wrapper(partial(data), data)
            data.__module__ = getattr(data, "__module__", "<nomodule>")
            data.__name__ = ".".join([getattr(data, "__name__", "<unnamed>"), name])
            data.__qualname__ = ".".join([getattr(data, "__qualname__", "<unnamed>"), name])
            self.__data[name] = _DynamicData(data)
        elif isinstance(data, pd.DataFrame):
            self.__data[name] = _StaticData(data)
        else:
            raise TypeError(
                f"Data source {name} must be a pandas DataFrame or function that returns a pandas DataFrame."
            )

    def __getitem__(self, name: DataSourceName) -> Union[_DynamicData, _StaticData]:
        """Returns the `_DynamicData` or `_StaticData` object associated with `name`."""
        try:
            return self.__data[name]
        except KeyError as exc:
            raise KeyError(f"Data source {name} does not exist.") from exc

    def _clear(self):
        # We do not actually call self.cache.clear() because (a) it would only work when self._cache_has_app is True,
        # which is not the case when e.g. Vizro._reset is called, and (b) because we do not want to accidentally
        # clear the cache if we eventually put a call to Vizro._reset inside Vizro(), since cache should be persisted
        # across server restarts.
        self.__init__()  # type: ignore[misc]

    @property
    def _cache_has_app(self) -> bool:
        """Detects whether self.cache.init_app has been called (as it is in Vizro) to attach a Flask app to the cache.

        Note that even NullCache needs to have an app attached before it can be "used". The only time the cache would
        not have an app attached is if the user tries to interact with the cache before Vizro() has been called.
        """
        cache_has_app = hasattr(self.cache, "app")
        if not cache_has_app and self.cache.config["CACHE_TYPE"] != "NullCache":
            # Try to prevent anyone from setting data_manager.cache after they've instantiated Vizro().
            # No need to emit a warning if the cache is left as NullCache; we only care about this if someone has
            # explicitly set a cache.
            # Eventually Vizro should probably have init_app method explicitly to clear this up so the order of
            # operations is more reliable. Alternatively we could just initialize cache at run time rather than build
            # time, which is what Flask-Caching is really designed for. This would require an extra step for users
            # though, since it could not go in Vizro.run() since that is not used in the case of gunicorn.
            warnings.warn(
                "Cache does not have Vizro app attached and so is not operational. Make sure you call "
                "Vizro() after you set data_manager.cache."
            )
        return cache_has_app


data_manager: DataManager = DataManager()



"""Ranking charts."""

import vizro.models as vm
import vizro.plotly.express as px

from pages._pages_utils import PAGE_GRID, gapminder, make_code_clipboard_from_py_file

ordered_bar = vm.Page(
    title="Ordered bar",
    path="ranking/ordered-bar",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is an ordered bar chart?

            An ordered bar chart displays bars with lengths proportional to their values, arranged in descending or
            ascending order. One axis shows the categories, and the other provides a value scale starting from zero.

            &nbsp;

            #### When should I use it?

            Use an ordered bar chart to help your audience compare sizes and identify patterns in categorical data,
            emphasizing the order of categories. This is ideal for showing rankings or priorities.
            Ensure clear labeling, especially with many bars, and consider using a legend or abbreviations with fuller
            descriptions below.
        """
        ),
        vm.Graph(
            figure=px.bar(
                gapminder.query(
                    "year == 2007 and country.isin(['United States', 'Pakistan', 'India', 'China', 'Indonesia'])"
                ).sort_values("pop"),
                x="pop",
                y="country",
                orientation="h",
            )
        ),
        make_code_clipboard_from_py_file("ordered_bar.py"),
    ],
)

ordered_column = vm.Page(
    title="Ordered column",
    path="ranking/ordered-column",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is an ordered column chart?

            An ordered column chart is a vertical bar chart where columns are arranged in descending or ascending order
            based on their values. The column lengths vary according to the categorical value they represent, with the
            scale on the y-axis starting from zero.

            &nbsp;

            #### When should I use it?

            Use an ordered column chart to help your audience compare sizes and identify patterns in categorical data,
            emphasizing the order of categories. This is ideal for showing rankings or progressions. Ensure clear
            labeling, especially with many columns, and consider using a legend or abbreviations with fuller
            descriptions below.
        """
        ),
        vm.Graph(
            figure=px.bar(
                gapminder.query(
                    "year == 2007 and country.isin(['United States', 'Pakistan', 'India', 'China', 'Indonesia'])"
                ).sort_values("pop"),
                y="pop",
                x="country",
            )
        ),
        make_code_clipboard_from_py_file("ordered_column.py"),
    ],
)


pages = [ordered_bar, ordered_column]



import re

import pytest

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm


class TestSharedValidators:
    def test_validate_min_length(self, model_with_layout):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            model_with_layout(title="Title", components=[])

    @pytest.mark.parametrize(
        "captured_callable, error_message",
        [
            (
                "standard_px_chart",
                "A callable of mode `graph` has been provided. Please wrap it inside `vm.Graph(figure=...)`",
            ),
            (
                "standard_ag_grid",
                "A callable of mode `ag_grid` has been provided. Please wrap it inside `vm.AgGrid(figure=...)`",
            ),
            (
                "standard_dash_table",
                "A callable of mode `table` has been provided. Please wrap it inside `vm.Table(figure=...)`",
            ),
            (
                "standard_kpi_card",
                "A callable of mode `figure` has been provided. Please wrap it inside `vm.Figure(figure=...)`",
            ),
        ],
    )
    def test_check_captured_callable(self, model_with_layout, captured_callable, error_message, request):
        with pytest.raises(ValidationError, match=re.escape(error_message)):
            model_with_layout(title="Title", components=[request.getfixturevalue(captured_callable)])

    def test_check_for_valid_component_types(self, model_with_layout):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "(allowed values: 'ag_grid', 'button', 'card', 'container', 'figure', 'graph', 'table', 'tabs')"
            ),
        ):
            model_with_layout(title="Page Title", components=[vm.Checklist()])



import pandas as pd
import plotly.graph_objects as go
import vizro.models as vm
import vizro.plotly.express as px
from plotly.subplots import make_subplots
from vizro import Vizro
from vizro.models.types import capture

gapminder = px.data.gapminder()


@capture("graph")
def column_and_line(data_frame: pd.DataFrame, x: str, y_column: str, y_line: str) -> go.Figure:
    """Creates a combined column and line chart using Plotly."""
    fig = make_subplots(specs=[[{"secondary_y": True}]])

    fig.add_trace(
        go.Bar(x=data_frame[x], y=data_frame[y_column], name=y_column),
        secondary_y=False,
    )

    fig.add_trace(
        go.Scatter(x=data_frame[x], y=data_frame[y_line], name=y_line),
        secondary_y=True,
    )

    fig.update_layout(
        xaxis={"type": "category", "title": x},
        yaxis={"tickmode": "sync", "title": y_column},
        yaxis2={"tickmode": "sync", "overlaying": "y", "title": y_line},
    )

    return fig


dashboard = vm.Dashboard(
    pages=[
        vm.Page(
            title="Column and line",
            components=[
                vm.Graph(
                    figure=column_and_line(
                        gapminder.query("country == 'Vietnam'"),
                        y_column="gdpPercap",
                        y_line="lifeExp",
                        x="year",
                    )
                )
            ],
        )
    ]
)
Vizro().build(dashboard).run()



"""Module containing the standard implementation of `dash_table.DataTable`."""

import pandas as pd
from dash import dash_table

from vizro.models.types import capture
from vizro.tables._utils import _set_defaults_nested


@capture("table")
def dash_data_table(data_frame: pd.DataFrame, **kwargs) -> dash_table.DataTable:
    """Standard `dash_table.DataTable` with sensible defaults to be used in [`Table`][vizro.models.Table].

    Examples
        Wrap inside `vm.Table` to use as a component inside `vm.Page` or `vm.Container`.
        >>> import vizro.models as vm
        >>> from vizro.table import dash_data_table
        >>> vm.Page(title="Page", components=[vm.Table(figure=dash_data_table(...))])

    """
    defaults = {
        "columns": [{"name": col, "id": col} for col in data_frame.columns],
        "style_as_list_view": True,
        "style_cell": {"position": "static"},
        "style_data": {"border_bottom": "1px solid var(--border-subtle-alpha-01)", "height": "40px"},
        "style_header": {
            "border_bottom": "1px solid var(--state-overlays-selected-hover)",
            "border_top": "1px solid var(--main-container-bg-color)",
            "height": "32px",
        },
        "style_data_conditional": [
            {
                "if": {"state": "active"},
                "backgroundColor": "var(--state-overlays-selected)",
                "border": "1px solid var(--state-overlays-selected)",
            }
        ],
    }
    kwargs = _set_defaults_nested(kwargs, defaults)
    return dash_table.DataTable(data=data_frame.to_dict("records"), **kwargs)



import pandas as pd
import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

funnel_data = pd.DataFrame(
    {"Stage": ["Leads", "Sales calls", "Follow-up", "Conversion", "Sales"], "Value": [10, 7, 4, 2, 1]}
)

page = vm.Page(
    title="Funnel",
    components=[vm.Graph(figure=px.funnel_area(funnel_data, names="Stage", values="Value"))],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Contains reusable data sets and constants."""

import logging
from pathlib import Path

import black
import pandas as pd
import vizro.plotly.express as px
from custom_components import CodeClipboard

# To disable logging info messages caused by black.format_str: https://github.com/psf/black/issues/2058
logging.getLogger("blib2to3").setLevel(logging.ERROR)


def make_code_clipboard_from_py_file(filepath: str):
    # Black doesn't yet have a Python API, so format_str might not work at some point in the future.
    # https://black.readthedocs.io/en/stable/faq.html#does-black-have-an-api
    filepath = Path(__file__).parents[1] / "pages/examples" / filepath
    return CodeClipboard(
        code=black.format_str(filepath.read_text(encoding="utf-8"), mode=black.Mode(line_length=80)),
        language="python",
    )


PAGE_GRID = [[0, 0, 0, 0, 0, 0, 0]] * 2 + [[1, 1, 1, 1, 2, 2, 2]] * 5

# DATA --------------------------------------------------------------
gapminder = px.data.gapminder()
iris = px.data.iris()
stocks = px.data.stocks()
tips = px.data.tips()

ages = pd.DataFrame(
    {
        "Age": ["0-19", "20-29", "30-39", "40-49", "50-59", ">=60"],
        "Male": [800, 2000, 4200, 5000, 2100, 800],
        "Female": [1000, 3000, 3500, 3800, 3600, 700],
    }
)
sankey_data = pd.DataFrame(
    {
        "Origin": [0, 1, 0, 2, 3, 3],
        "Destination": [2, 3, 3, 4, 4, 5],
        "Value": [8, 4, 2, 8, 4, 2],
    }
)

funnel_data = pd.DataFrame(
    {"Stage": ["Leads", "Sales calls", "Follow-up", "Conversion", "Sales"], "Value": [10, 7, 4, 2, 1]}
)

stepped_line_data = pd.DataFrame(
    {
        "year": [1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003],
        "rate": [0.10, 0.12, 0.15, 0.13, 0.14, 0.13, 0.14, 0.16, 0.15],
    }
)



from __future__ import annotations

from typing import List, Literal, Union

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype, is_numeric_dtype

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

from vizro._constants import FILTER_ACTION_PREFIX
from vizro.actions import _filter
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models import Action, VizroBaseModel
from vizro.models._components.form import (
    Checklist,
    DatePicker,
    Dropdown,
    RadioItems,
    RangeSlider,
    Slider,
)
from vizro.models._models_utils import _log_call
from vizro.models.types import MultiValueType, SelectorType

# Ideally we might define these as NumericalSelectorType = Union[RangeSlider, Slider] etc., but that will not work
# with isinstance checks.
# First entry in each tuple is the default selector for that column type.
SELECTORS = {
    "numerical": (RangeSlider, Slider),
    "categorical": (Dropdown, Checklist, RadioItems),
    "temporal": (DatePicker,),
}

# This disallowed selectors for each column type map is based on the discussion at the following link:
# See https://github.com/mckinsey/vizro/pull/319#discussion_r1524888171
DISALLOWED_SELECTORS = {
    "numerical": SELECTORS["temporal"],
    "temporal": SELECTORS["numerical"],
    "categorical": SELECTORS["numerical"] + SELECTORS["temporal"],
}


def _filter_between(series: pd.Series, value: Union[List[float], List[str]]) -> pd.Series:
    if is_datetime64_any_dtype(series):
        # Each value will always have time 00:00:00. In order for the filter to include all times during
        # the end date value[1] we need to remove the time part of every value in series so that it's 00:00:00.
        value = pd.to_datetime(value)
        series = pd.to_datetime(series.dt.date)
    return series.between(value[0], value[1], inclusive="both")


def _filter_isin(series: pd.Series, value: MultiValueType) -> pd.Series:
    if is_datetime64_any_dtype(series):
        # Value will always have time 00:00:00. In order for the filter to include all times during
        # the end date value we need to remove the time part of every value in series so that it's 00:00:00.
        value = pd.to_datetime(value)
        series = pd.to_datetime(series.dt.date)
    return series.isin(value)


class Filter(VizroBaseModel):
    """Filter the data supplied to `targets` on the [`Page`][vizro.models.Page].

    Examples:
        >>> print(repr(Filter(column="species")))

    Args:
        type (Literal["filter"]): Defaults to `"filter"`.
        column (str): Column of `DataFrame` to filter.
        targets (List[ModelID]): Target component to be affected by filter. If none are given then target all components
            on the page that use `column`.
        selector (SelectorType): See [SelectorType][vizro.models.types.SelectorType]. Defaults to `None`.

    """

    type: Literal["filter"] = "filter"
    column: str = Field(..., description="Column of DataFrame to filter.")
    targets: List[ModelID] = Field(
        [],
        description="Target component to be affected by filter. "
        "If none are given then target all components on the page that use `column`.",
    )
    selector: SelectorType = None
    _column_type: Literal["numerical", "categorical", "temporal"] = PrivateAttr()

    @validator("targets", each_item=True)
    def check_target_present(cls, target):
        if target not in model_manager:
            raise ValueError(f"Target {target} not found in model_manager.")
        return target

    @_log_call
    def pre_build(self):
        self._set_targets()
        self._set_column_type()
        self._set_selector()
        self._validate_disallowed_selector()
        self._set_numerical_and_temporal_selectors_values()
        self._set_categorical_selectors_options()
        self._set_actions()

    @_log_call
    def build(self):
        return self.selector.build()

    def _set_targets(self):
        if not self.targets:
            for component_id in model_manager._get_page_model_ids_with_figure(
                page_id=model_manager._get_model_page_id(model_id=ModelID(str(self.id)))
            ):
                # TODO: consider making a helper method in data_manager or elsewhere to reduce this operation being
                #  duplicated across Filter so much, and/or consider storing the result to avoid repeating it.
                #  Need to think about this in connection with how to update filters on the fly and duplicated calls
                #  issue outlined in https://github.com/mckinsey/vizro/pull/398#discussion_r1559120849.
                data_source_name = model_manager[component_id]["data_frame"]
                data_frame = data_manager[data_source_name].load()
                if self.column in data_frame.columns:
                    self.targets.append(component_id)
            if not self.targets:
                raise ValueError(f"Selected column {self.column} not found in any dataframe on this page.")

    def _set_column_type(self):
        data_source_name = model_manager[self.targets[0]]["data_frame"]
        data_frame = data_manager[data_source_name].load()

        if is_numeric_dtype(data_frame[self.column]):
            self._column_type = "numerical"
        elif is_datetime64_any_dtype(data_frame[self.column]):
            self._column_type = "temporal"
        else:
            self._column_type = "categorical"

    def _set_selector(self):
        self.selector = self.selector or SELECTORS[self._column_type][0]()
        self.selector.title = self.selector.title or self.column.title()

    def _validate_disallowed_selector(self):
        if isinstance(self.selector, DISALLOWED_SELECTORS.get(self._column_type, ())):
            raise ValueError(
                f"Chosen selector {self.selector.type} is not compatible "
                f"with {self._column_type} column '{self.column}'. "
            )

    def _set_numerical_and_temporal_selectors_values(self):
        # If the selector is a numerical or temporal selector, and the min and max values are not set, then set them
        # N.B. All custom selectors inherit from numerical or temporal selector should also pass this check
        if isinstance(self.selector, SELECTORS["numerical"] + SELECTORS["temporal"]):
            min_values = []
            max_values = []
            for target_id in self.targets:
                data_source_name = model_manager[target_id]["data_frame"]
                data_frame = data_manager[data_source_name].load()
                min_values.append(data_frame[self.column].min())
                max_values.append(data_frame[self.column].max())

            if not (
                is_numeric_dtype(pd.Series(min_values))
                and is_numeric_dtype(pd.Series(max_values))
                or is_datetime64_any_dtype(pd.Series(min_values))
                and is_datetime64_any_dtype(pd.Series(max_values))
            ):
                raise ValueError(
                    f"Inconsistent types detected in the shared data column '{self.column}' for targeted charts "
                    f"{self.targets}. Please ensure that the data column contains the same data type across all "
                    f"targeted charts."
                )

            if self.selector.min is None:
                self.selector.min = min(min_values)
            if self.selector.max is None:
                self.selector.max = max(max_values)

    def _set_categorical_selectors_options(self):
        # If the selector is a categorical selector, and the options are not set, then set them
        # N.B. All custom selectors inherit from categorical selector should also pass this check
        if isinstance(self.selector, SELECTORS["categorical"]) and not self.selector.options:
            options = set()
            for target_id in self.targets:
                data_source_name = model_manager[target_id]["data_frame"]
                data_frame = data_manager[data_source_name].load()
                options |= set(data_frame[self.column])

            self.selector.options = sorted(options)

    def _set_actions(self):
        if not self.selector.actions:
            if isinstance(self.selector, RangeSlider) or (
                isinstance(self.selector, DatePicker) and self.selector.range
            ):
                filter_function = _filter_between
            else:
                filter_function = _filter_isin

            self.selector.actions = [
                Action(
                    id=f"{FILTER_ACTION_PREFIX}_{self.id}",
                    function=_filter(filter_column=self.column, targets=self.targets, filter_function=filter_function),
                )
            ]



import logging
from typing import Dict, List, Literal

from dash import ClientsideFunction, Input, Output, State, clientside_callback, ctx, dcc
from dash.exceptions import MissingCallbackContextException
from plotly import graph_objects as go

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

import pandas as pd

from vizro.actions._actions_utils import CallbackTriggerDict, _get_component_actions
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components._components_utils import _process_callable_data_frame
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Graph(VizroBaseModel):
    """Wrapper for `dcc.Graph` to visualize charts in dashboard.

    Args:
        type (Literal["graph"]): Defaults to `"graph"`.
        figure (CapturedCallable): Function that returns a graph.
            See `CapturedCallable`][vizro.models.types.CapturedCallable].
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.

    """

    type: Literal["graph"] = "graph"
    figure: CapturedCallable = Field(
        ..., import_path="vizro.plotly.express", mode="graph", description="Function that returns a plotly `go.Figure`"
    )
    actions: List[Action] = []

    # Component properties for actions and interactions
    _output_component_property: str = PrivateAttr("figure")

    # Validators
    _set_actions = _action_validator_factory("clickData")
    _validate_callable = validator("figure", allow_reuse=True)(_process_callable_data_frame)

    # Convenience wrapper/syntactic sugar.
    def __call__(self, **kwargs):
        # This default value is not actually used anywhere at the moment since __call__ is always used with data_frame
        # specified. It's here to match Table and AgGrid and because we might want to use __call__ more in future.
        # If the functionality of process_callable_data_frame moves to CapturedCallable then this would move there too.
        kwargs.setdefault("data_frame", data_manager[self["data_frame"]].load())
        fig = self.figure(**kwargs)
        fig = self._optimise_fig_layout_for_dashboard(fig)

        # Apply the template vizro_dark or vizro_light by setting fig.layout.template. This is exactly the same as
        # what the clientside update_graph_theme callback does, and it would be nice if we could just use that by
        # including Input(self.id, "figure") as input for that callback, but doing so leads to a small flicker between
        # completion of this serverside callback and starting that clientside callbacks.
        # Note that this does not fully set the template for plotly.express figures. Doing this post-fig creation update
        # relies on the fact that we have already set pio.templates.default before the self.figure call
        # above (but not with the _pio_templates_default context manager surrounding the above self.figure call,
        # since that would alter global state).
        # Possibly we should pass through the theme selector as an argument `template` in __call__ rather than fetching
        # it from ctx here. Remember that passing it as self.figure(template) is not helpful though, because custom
        # graph figures don't need a template argument, and the clientside them selector callback would override this
        # anyway.
        # Possibly we should enforce that __call__ can only be used within the context of a callback, but it's easy
        # to just swallow up the error here as it doesn't cause any problems.
        try:
            # At the moment theme_selector is always present so this if statement is redundant, but possibly in
            # future we'll have callbacks that do Graph.__call__() without theme_selector set.
            if "theme_selector" in ctx.args_grouping.get("external", {}):
                theme_selector_checked = ctx.args_grouping["external"]["theme_selector"]["value"]
                fig.layout.template = "vizro_light" if theme_selector_checked else "vizro_dark"
        except MissingCallbackContextException:
            logger.info("fig.update_layout called outside of callback context.")
        return fig

    # Convenience wrapper/syntactic sugar.
    def __getitem__(self, arg_name: str):
        # See figure implementation for more details.
        if arg_name == "type":
            return self.type
        return self.figure[arg_name]

    # Interaction methods
    @property
    def _filter_interaction_input(self):
        """Required properties when using pre-defined `filter_interaction`."""
        return {
            "clickData": State(component_id=self.id, component_property="clickData"),
            "modelID": State(component_id=self.id, component_property="id"),  # required, to determine triggered model
        }

    def _filter_interaction(
        self, data_frame: pd.DataFrame, target: str, ctd_filter_interaction: Dict[str, CallbackTriggerDict]
    ) -> pd.DataFrame:
        """Function to be carried out for pre-defined `filter_interaction`."""
        # data_frame is the DF of the target, ie the data to be filtered, hence we cannot get the DF from this model
        ctd_click_data = ctd_filter_interaction["clickData"]
        if not ctd_click_data["value"]:
            return data_frame

        source_graph_id: ModelID = ctd_click_data["id"]
        source_graph_actions = _get_component_actions(model_manager[source_graph_id])
        try:
            custom_data_columns = model_manager[source_graph_id]["custom_data"]
        except KeyError as exc:
            raise KeyError(
                f"Missing 'custom_data' for the source graph with id {source_graph_id}. "
                "Ensure that `custom_data` is an argument of the custom chart function, and that the relevant entry is "
                "then passed to the underlying plotly function. When configuring the custom chart in `vm.Graph`, "
                "ensure that `custom_data` is passed. Example usage: "
                "vm.Graph(figure=my_custom_chart(df, custom_data=['column_1'], actions=[...]))"
            ) from exc

        customdata = ctd_click_data["value"]["points"][0]["customdata"]

        for action in source_graph_actions:
            if action.function._function.__name__ != "filter_interaction" or target not in action.function["targets"]:
                continue
            for custom_data_idx, column in enumerate(custom_data_columns):
                data_frame = data_frame[data_frame[column].isin([customdata[custom_data_idx]])]

        return data_frame

    @staticmethod
    def _optimise_fig_layout_for_dashboard(fig):
        """Post layout updates to visually enhance charts used inside dashboard."""
        # Reduce `margin_t` if no title is provided and `margin_t` is not explicitly set.
        if fig.layout.margin.t is None and fig.layout.title.text is None:
            fig.update_layout(margin_t=24, title_pad_l=0, title_pad_r=0, margin_l=24)

        # Reduce `title_pad_t` if no subtitle is provided and `title_pad_t` is not explicitly set.
        if fig.layout.title.pad.t is None and fig.layout.title.text and "<br>" not in fig.layout.title.text:
            fig.update_layout(title_pad_t=7)

        if fig.layout.title.pad.l is None:
            fig.update_layout(title_pad_l=0)

        if fig.layout.title.pad.r is None:
            fig.update_layout(title_pad_r=0)

        if fig.layout.margin.l is None:
            fig.update_layout(margin_l=24)

        return fig

    @_log_call
    def build(self):
        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="update_graph_theme"),
            # Output here to ensure that the callback is only triggered if the graph exists on the currently open page.
            output=[Output(self.id, "figure")],
            inputs=[
                Input("theme_selector", "checked"),
                State("vizro_themes", "data"),
                State(self.id, "id"),
            ],
            prevent_initial_call=True,
        )

        # The empty figure here is just a placeholder designed to be replaced by the actual figure when the filters
        # etc. are applied. It only appears on the screen for a brief instant, but we need to make sure it's
        # transparent and has no axes so it doesn't draw anything on the screen which would flicker away when the
        # graph callback is executed to make the dcc.Loading icon appear.
        return dcc.Loading(
            dcc.Graph(
                id=self.id,
                figure=go.Figure(
                    layout={
                        "paper_bgcolor": "rgba(0,0,0,0)",
                        "plot_bgcolor": "rgba(0,0,0,0)",
                        "xaxis": {"visible": False},
                        "yaxis": {"visible": False},
                    }
                ),
                config={"autosizable": True, "frameMargins": 0, "responsive": True},
                className="chart_container",
            ),
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )



"""Part-to-whole charts."""

import vizro.models as vm
import vizro.plotly.express as px

from pages._pages_utils import PAGE_GRID, funnel_data, gapminder, make_code_clipboard_from_py_file, tips

pie = vm.Page(
    title="Pie",
    path="part-to-whole/pie",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a pie chart?

            A pie chart is a circular chart divided into segments to show proportions and percentages between
            categories. The circle represents the whole.

            &nbsp;

            #### When should I use it?

            Use the pie chart when you need to show your audience a quick view of how data is distributed
            proportionately, with percentages highlighted. The different values you present must add up to a total and
            equal 100%.

            The downsides are that pie charts tend to occupy more space than other charts, they don't
            work well with more than a few values because labeling small segments is challenging, and it can be
            difficult to accurately compare the sizes of the segments.
        """
        ),
        vm.Graph(
            figure=px.pie(
                tips,
                values="tip",
                names="day",
            )
        ),
        make_code_clipboard_from_py_file("pie.py"),
    ],
)

donut = vm.Page(
    title="Donut",
    path="part-to-whole/donut",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a donut chart?

            A donut chart looks like a pie chart, but has a blank space in the center which may contain additional
            information.

            &nbsp;

            #### When should I use it?

            A donut chart can be used in place of a pie chart, particularly when you are short of space or have extra
            information you would like to share about the data. It may also be more effective if you wish your audience
            to focus on the length of the arcs of the sections instead of the proportions of the segment sizes.
        """
        ),
        vm.Graph(
            figure=px.pie(
                tips,
                values="tip",
                names="day",
                hole=0.4,
            )
        ),
        make_code_clipboard_from_py_file("donut.py"),
    ],
)

treemap = vm.Page(
    title="Treemap",
    path="part-to-whole/treemap",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

                #### What is a treemap?

                A treemap shows hierarchical data arranged as a set of nested rectangles: rectangles are sized
                proportionately to the quantity they represent, combined together to form larger parent category
                rectangles.

                &nbsp;

                #### When should I use it?

                It's helpful to use a treemap when you wish to display hierarchical part-to-whole relationships. You can
                compare groups and single elements nested within them. Consider using them instead of Pie charts when
                you have a higher number of categories. Treemaps are very compact and allow audiences to get a quick
                overview of the data.
            """
        ),
        vm.Graph(
            figure=px.treemap(
                gapminder.query("year == 2007"),
                path=[px.Constant("world"), "continent", "country"],
                values="pop",
                color="lifeExp",
            )
        ),
        make_code_clipboard_from_py_file("treemap.py"),
    ],
)

stacked_bar = vm.Page(
    title="Stacked bar",
    path="part-to-whole/stacked-bar",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a stacked bar chart?

            A stacked bar chart displays bars divided into segments, with each segment's length proportional to the
            value it represents. One axis shows the categories being compared, while the other provides a value scale
            starting from zero. The segments within each bar are stacked on top of each other, allowing for a cumulative
            comparison.

            &nbsp;

            #### When should I use it?

            Use a stacked bar chart to help your audience compare the total sizes of categories as well as the
            individual components within those categories. This chart type is ideal for visualizing part-to-whole
            relationships and identifying patterns within categorical data. Ensure clear labeling for each segment,
            especially when there are many categories, and consider using a legend or abbreviations with fuller
            descriptions below.
        """
        ),
        vm.Graph(figure=px.histogram(tips, y="sex", x="total_bill", color="day", orientation="h")),
        make_code_clipboard_from_py_file("stacked_bar.py"),
    ],
)

stacked_column = vm.Page(
    title="Stacked column",
    path="part-to-whole/stacked-column",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a stacked column chart?

            A stacked column chart displays columns divided into segments, with each segment's height proportional to
            the value it represents. One axis shows the categories being compared, while the other provides a value
            scale starting from zero. The segments within each column are stacked on top of each other, allowing for a
            cumulative comparison.

            &nbsp;

            #### When should I use it?

            Use a stacked column chart to help your audience compare the total sizes of categories as well as the
            individual components within those categories. This chart type is ideal for visualizing part-to-whole
            relationships and identifying patterns within categorical data. Ensure clear labeling for each segment,
            especially when there are many categories, and consider using a legend or abbreviations with fuller
            descriptions below.
        """
        ),
        vm.Graph(figure=px.histogram(tips, x="sex", y="total_bill", color="day")),
        make_code_clipboard_from_py_file("stacked_bar.py"),
    ],
)

funnel = vm.Page(
    title="Funnel",
    path="part-to-whole/funnel",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a funnel chart?

            A funnel area chart is a type of data visualization that represents stages in a process, with the size of
            each area proportional to its value. The chart typically narrows as it progresses, visually depicting the
            reduction in numbers through each stage. One axis represents the stages of the process, while the other axis
            indicates the values or quantities at each stage.

            &nbsp;

            #### When should I use it?

            Use a funnel area chart to help your audience understand and compare the progression of data through
            different stages of a process. This chart type is particularly effective for visualizing conversion rates,
            sales processes, or any sequential data where you want to highlight drop-offs or reductions between stages.
        """
        ),
        vm.Graph(figure=px.funnel_area(funnel_data, names="Stage", values="Value")),
        make_code_clipboard_from_py_file("funnel.py"),
    ],
)


pages = [donut, pie, treemap, stacked_bar, stacked_column, funnel]



# TODO: eventually deduplicate page generation into a function rather than copying and pasting across files?
# TODO: think about the best way to do code examples, e.g.
# - do we want full dashboard example or plot-only example?
# - or both? Could be done using a toggle switch or multiple tabs.
# - a link to py.cafe showing the dashboard code?



"""Dev app to try things out."""

import pandas as pd
import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro
from vizro.actions import filter_interaction
from vizro.figures import kpi_card
from vizro.managers import data_manager
from vizro.models.types import capture
from vizro.tables import dash_ag_grid, dash_data_table

data = pd.DataFrame(
    {
        "Column 1": [1, 2, 3, 4, 5, 6],
        "Column 2": ["A", "B", "C", "VeryLongStringInputCell_VeryLongStringInputCell", "D", "E"],
        "Column 3": [10.5, 20.1, 30.2, 40.3, 50.4, 60.5],
    }
)

iris_dataset = px.data.iris()
giant_dataset = pd.concat([iris_dataset] * 10000, ignore_index=True)


def load_data():
    """Load data."""
    return data


data_manager["my_data"] = load_data


@capture("ag_grid")
def custom_dash_ag_grid(data_frame, **kwargs):
    """Custom AgGrid."""
    # print(f"dash_ag_grid -> len: {len(data_frame)}")
    return dash_ag_grid(data_frame, **kwargs)()


@capture("table")
def custom_dash_data_table(data_frame, **kwargs):
    """Custom DataTable."""
    # print(f"dash_data_table -> len: {len(data_frame)}")
    return dash_data_table(data_frame, **kwargs)()


@capture("graph")
def custom_px_scatter(data_frame, **kwargs):
    """Custom Scatter plot."""
    # print(f"graph -> len: {len(data_frame)}")
    return px.scatter(data_frame, **kwargs)


@capture("figure")
def custom_kpi_card(data_frame, **kwargs):
    """Custom KPI card."""
    # print(f"kpi_card -> len: {len(data_frame)}")
    return kpi_card(data_frame, **kwargs)()


page_grid = vm.Page(
    title="Example Page",
    layout=vm.Layout(
        grid=[
            [0, 1],
            [2, 3],
        ]
    ),
    components=[
        vm.AgGrid(
            id="outer_ag_grid_id",
            figure=custom_dash_ag_grid(
                id="inner_ag_grid_id",
                data_frame="my_data",
                columnDefs=[{"field": col, "checkboxSelection": True, "filter": True} for col in data.columns],
                columnSize="autoSize",
                defaultColDef={"resizable": True},
                persistence=True,
                persistence_type="local",
                persisted_props=["selectedRows", "filterModel"],
                dashGridOptions={
                    "rowSelection": "multiple",
                    "suppressRowClickSelection": True,
                    # Turn pagination on to see results:
                    # "pagination": True,
                    # "paginationPageSize": 3,
                },
            ),
            actions=[
                vm.Action(
                    function=filter_interaction(targets=["graph_id"]),
                )
            ],
        ),
        vm.Table(
            id="outer_table_id",
            figure=custom_dash_data_table(
                id="inner_table_id",
                data_frame="my_data",
                row_selectable="multi",
                filter_action="native",
                persistence=True,
                persistence_type="local",
                # Turn pagination on to see results:
                # page_action="native",
                # page_size=3,
            ),
        ),
        vm.Graph(
            id="graph_id",
            figure=custom_px_scatter(data_frame="my_data", x="Column 1", y="Column 3"),
        ),
        vm.Figure(
            id="figure_id",
            figure=custom_kpi_card(data_frame="my_data", value_column="Column 1"),
        ),
    ],
    controls=[vm.Filter(column="Column 1", selector=vm.RangeSlider(step=1))],
)

columnDefs = [{"field": "petal_length"}]

dashboard = vm.Dashboard(
    pages=[
        vm.Page(
            title="Page_1",
            components=[vm.Card(text="Dummy page just for testing")],
        ),
        page_grid,
        vm.Page(
            title="Page_3",
            components=[vm.AgGrid(figure=dash_ag_grid(data_frame=giant_dataset, columnDefs=columnDefs))],
        ),
    ]
)


if __name__ == "__main__":
    Vizro().build(dashboard).run(debug=True)



from typing import Literal

from dash import dcc, html

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

from vizro.managers import data_manager
from vizro.models import VizroBaseModel
from vizro.models._components._components_utils import _process_callable_data_frame
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable


class Figure(VizroBaseModel):
    """Creates a figure-like object that can be displayed in the dashboard and is reactive to controls.

    Args:
        type (Literal["figure"]): Defaults to `"figure"`.
        figure (CapturedCallable): Function that returns a figure-like object. See [`vizro.figures`][vizro.figures].

    """

    type: Literal["figure"] = "figure"
    figure: CapturedCallable = Field(
        import_path="vizro.figures",
        mode="figure",
        description="Function that returns a figure-like object.",
    )

    # Component properties for actions and interactions
    _output_component_property: str = PrivateAttr("children")

    # Validators
    _validate_callable = validator("figure", allow_reuse=True, always=True)(_process_callable_data_frame)

    def __call__(self, **kwargs):
        kwargs.setdefault("data_frame", data_manager[self["data_frame"]].load())
        figure = self.figure(**kwargs)
        return figure

    def __getitem__(self, arg_name: str):
        # pydantic discriminated union validation seems to try Figure["type"], which throws an error unless we
        # explicitly redirect it to the correct attribute.
        if self.figure:
            if arg_name == "type":
                return self.type
            return self.figure[arg_name]

    @_log_call
    def build(self):
        return dcc.Loading(
            # Refer to the vm.AgGrid build method for details on why we return the
            # html.Div(id=self.id) instead of actual figure object with the original data_frame.
            # Optimally, we would like to provide id=self.id directly here such that we can target the CSS
            # of the children via ID as well, but the `id` doesn't seem to be passed on to the loading component.
            # This limitation is handled with this PR -> https://github.com/plotly/dash/pull/2888.
            # The PR is merged but is not released yet. Once it is released, we can try to refactor the following code.
            # In the meantime, we are adding an extra html.div here.
            html.Div(id=self.id, className="figure-container"),
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )



from typing import List, Literal, Optional, Union

import dash_mantine_components as dmc
from dash import ClientsideFunction, Input, Output, State, clientside_callback, dcc, html

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator


import datetime
from datetime import date

import dash_bootstrap_components as dbc

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components.form._form_utils import validate_date_picker_range, validate_max, validate_range_value


class DatePicker(VizroBaseModel):
    """Temporal single/range option selector `DatePicker`.

    Can be provided to [`Filter`][vizro.models.Filter] or [`Parameter`][vizro.models.Parameter].
    Based on the underlying [`dmc.DatePicker`](https://www.dash-mantine-components.com/components/datepicker) or
    [`dmc.DateRangePicker`](https://www.dash-mantine-components.com/components/datepicker#daterangepicker).

    Args:
        type (Literal["date_picker"]): Defaults to `"date_picker"`.
        min (Optional[date]): Start date for date picker. Defaults to `None`.
        max (Optional[date]): End date for date picker. Defaults to `None`.
        value (Union[List[date], date]): Default date/dates for date picker. Defaults to `None`.
        title (str): Title to be displayed. Defaults to `""`.
        range (bool): Boolean flag for displaying range picker. Default to `True`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.

    """

    type: Literal["date_picker"] = "date_picker"
    min: Optional[date] = Field(None, description="Start date for date picker.")
    max: Optional[date] = Field(None, description="End date for date picker.")
    value: Optional[Union[List[date], date]] = Field(None, description="Default date for date picker")
    title: str = Field("", description="Title to be displayed.")
    range: bool = Field(True, description="Boolean flag for displaying range picker.")
    actions: List[Action] = []

    _input_property: str = PrivateAttr("value")
    _set_actions = _action_validator_factory("value")

    # Re-used validators
    _validate_value = validator("value", allow_reuse=True)(validate_range_value)
    _validate_max = validator("max", allow_reuse=True)(validate_max)
    _validate_range = validator("range", allow_reuse=True, always=True)(validate_date_picker_range)

    def build(self):
        init_value = self.value or ([self.min, self.max] if self.range else self.min)  # type: ignore[list-item]
        date_range_picker_kwargs = {"allowSingleDateInRange": True} if self.range else {}

        output = [
            Output(self.id, "value"),
            Output(f"{self.id}_input_store", "data"),
        ]
        inputs = [
            Input(self.id, "value"),
            State(f"{self.id}_input_store", "data"),
        ]

        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="update_date_picker_values"),
            output=output,
            inputs=inputs,
        )
        # clientside callback is required as a workaround when the date-picker is overflowing its parent container
        # if there is not enough space. Caused by another workaround for this issue:
        # https://github.com/snehilvj/dash-mantine-components/issues/219
        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="update_date_picker_position"),
            output=Output(self.id, "dropdownPosition"),
            inputs=Input(self.id, "n_clicks"),
        )

        date_picker_class = dmc.DateRangePicker if self.range else dmc.DatePicker

        # dropdownPosition must be set to bottom-start as a workaround for issue:
        # https://github.com/snehilvj/dash-mantine-components/issues/219
        # clearable must be set to False as a workaround for issue:
        # https://github.com/snehilvj/dash-mantine-components/issues/212
        # maxDate must be increased by one day, and later on disabledDates must be set as maxDate + 1 day
        # as a workaround for issue: https://github.com/snehilvj/dash-mantine-components/issues/230
        date_picker = date_picker_class(
            id=self.id,
            minDate=self.min,
            value=init_value,
            maxDate=self.max + datetime.timedelta(days=1) if self.max else None,
            persistence=True,
            persistence_type="session",
            dropdownPosition="bottom-start",
            clearable=False,
            disabledDates=self.max + datetime.timedelta(days=1) if self.max else None,
            className="datepicker",
            **date_range_picker_kwargs,
        )

        return html.Div(
            children=[
                dbc.Label(children=self.title, html_for=self.id) if self.title else None,
                date_picker,
                dcc.Store(id=f"{self.id}_input_store", storage_type="session", data=init_value),
            ],
            className="selector_container",
        )



"""Unit tests for vizro.models.AgGrid."""

import re

import pytest
from asserts import assert_component_equal
from dash import dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro.managers import data_manager
from vizro.models._action._action import Action
from vizro.tables import dash_ag_grid


@pytest.fixture
def dash_ag_grid_with_id():
    return dash_ag_grid(id="underlying_table_id", data_frame=px.data.gapminder())


@pytest.fixture
def dash_ag_grid_with_arguments():
    return dash_ag_grid(data_frame=px.data.gapminder(), defaultColDef={"resizable": False, "sortable": False})


@pytest.fixture
def dash_ag_grid_with_str_dataframe():
    return dash_ag_grid(data_frame="gapminder")


class TestAgGridInstantiation:
    def test_create_graph_mandatory_only(self, standard_ag_grid):
        ag_grid = vm.AgGrid(figure=standard_ag_grid)

        assert hasattr(ag_grid, "id")
        assert ag_grid.type == "ag_grid"
        assert ag_grid.figure == standard_ag_grid
        assert ag_grid.actions == []

    @pytest.mark.parametrize("id", ["id_1", "id_2"])
    def test_create_ag_grid_mandatory_and_optional(self, standard_ag_grid, id):
        ag_grid = vm.AgGrid(figure=standard_ag_grid, id=id, actions=[])

        assert ag_grid.id == id
        assert ag_grid.type == "ag_grid"
        assert ag_grid.figure == standard_ag_grid

    def test_mandatory_figure_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.AgGrid()

    def test_captured_callable_invalid(self, standard_go_chart):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "Invalid CapturedCallable. Supply a function imported from vizro.tables or "
                "defined with decorator @capture('ag_grid')."
            ),
        ):
            vm.AgGrid(figure=standard_go_chart)

    def test_captured_callable_wrong_mode(self, standard_dash_table):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "CapturedCallable was defined with @capture('table') rather than @capture('ag_grid') and so "
                "is not compatible with the model."
            ),
        ):
            vm.AgGrid(figure=standard_dash_table)

    def test_is_model_inheritable(self, standard_ag_grid):
        class MyAgGrid(vm.AgGrid):
            pass

        my_ag_grid = MyAgGrid(figure=standard_ag_grid)

        assert hasattr(my_ag_grid, "id")
        assert my_ag_grid.type == "ag_grid"
        assert my_ag_grid.figure == standard_ag_grid
        assert my_ag_grid.actions == []

    def test_set_action_via_validator(self, standard_ag_grid, identity_action_function):
        ag_grid = vm.AgGrid(figure=standard_ag_grid, actions=[Action(function=identity_action_function())])
        actions_chain = ag_grid.actions[0]
        assert actions_chain.trigger.component_property == "cellClicked"


class TestDunderMethodsAgGrid:
    def test_getitem_known_args(self, dash_ag_grid_with_arguments):
        ag_grid = vm.AgGrid(figure=dash_ag_grid_with_arguments)
        assert ag_grid["defaultColDef"] == {"resizable": False, "sortable": False}
        assert ag_grid["type"] == "ag_grid"

    def test_getitem_unknown_args(self, standard_ag_grid):
        ag_grid = vm.AgGrid(figure=standard_ag_grid)
        with pytest.raises(KeyError):
            ag_grid["unknown_args"]

    def test_underlying_id_is_auto_generated(self, standard_ag_grid):
        ag_grid = vm.AgGrid(id="text_ag_grid", figure=standard_ag_grid)
        ag_grid.pre_build()
        # ag_grid() is the same as ag_grid.__call__()
        assert ag_grid().id == "__input_text_ag_grid"

    def test_underlying_id_is_provided(self, dash_ag_grid_with_id):
        ag_grid = vm.AgGrid(figure=dash_ag_grid_with_id)
        ag_grid.pre_build()
        # ag_grid() is the same as ag_grid.__call__()
        assert ag_grid().id == "underlying_table_id"


class TestAttributesAgGrid:
    # Testing at this low implementation level as mocking callback contexts skips checking for creation of these objects
    def test_ag_grid_filter_interaction_attributes(self, ag_grid_with_id):
        ag_grid = vm.AgGrid(figure=ag_grid_with_id, title="Gapminder", actions=[])
        ag_grid.pre_build()
        assert hasattr(ag_grid, "_filter_interaction_input")
        assert "modelID" in ag_grid._filter_interaction_input


class TestProcessAgGridDataFrame:
    def test_process_figure_data_frame_str_df(self, dash_ag_grid_with_str_dataframe, gapminder):
        data_manager["gapminder"] = gapminder
        ag_grid = vm.AgGrid(id="ag_grid", figure=dash_ag_grid_with_str_dataframe)
        assert data_manager[ag_grid["data_frame"]].load().equals(gapminder)

    def test_process_figure_data_frame_df(self, standard_ag_grid, gapminder):
        ag_grid = vm.AgGrid(id="ag_grid", figure=standard_ag_grid)
        assert data_manager[ag_grid["data_frame"]].load().equals(gapminder)


class TestPreBuildAgGrid:
    def test_pre_build_no_actions_no_underlying_ag_grid_id(self, standard_ag_grid):
        ag_grid = vm.AgGrid(id="text_ag_grid", figure=standard_ag_grid)
        ag_grid.pre_build()

        assert ag_grid._input_component_id == "__input_text_ag_grid"

    def test_pre_build_actions_underlying_ag_grid_id(self, ag_grid_with_id, filter_interaction_action):
        ag_grid = vm.AgGrid(id="text_ag_grid", figure=ag_grid_with_id, actions=[filter_interaction_action])
        ag_grid.pre_build()
        assert ag_grid._input_component_id == "underlying_ag_grid_id"


class TestBuildAgGrid:
    def test_ag_grid_build_mandatory_only(self, standard_ag_grid, gapminder):
        ag_grid = vm.AgGrid(id="text_ag_grid", figure=standard_ag_grid)
        ag_grid.pre_build()
        ag_grid = ag_grid.build()
        expected_ag_grid = dcc.Loading(
            [
                None,
                html.Div(
                    id="text_ag_grid",
                    children=[html.Div(id="__input_text_ag_grid")],
                    className="table-container",
                ),
            ],
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )

        assert_component_equal(ag_grid, expected_ag_grid)

    def test_ag_grid_build_with_underlying_id(self, ag_grid_with_id_and_conf, filter_interaction_action, gapminder):
        ag_grid = vm.AgGrid(id="text_ag_grid", figure=ag_grid_with_id_and_conf, actions=[filter_interaction_action])
        ag_grid.pre_build()
        ag_grid = ag_grid.build()

        expected_ag_grid = dcc.Loading(
            [
                None,
                html.Div(
                    id="text_ag_grid",
                    children=[html.Div(id="underlying_ag_grid_id")],
                    className="table-container",
                ),
            ],
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )

        assert_component_equal(ag_grid, expected_ag_grid)



"""Defines chart groups."""

import itertools
from dataclasses import dataclass
from typing import List

import pages.correlation
import pages.deviation
import pages.distribution
import pages.flow
import pages.magnitude
import pages.part_to_whole
import pages.ranking
import pages.spatial
import pages.time
import vizro.models as vm


class IncompletePage:
    """Fake vm.Page-like class.

    This has the properties required to make it function sufficiently like a page when generating the navigation cards.
    Only the title is configurable; path is fixed to "".
    """

    def __init__(self, title):  # noqa: D107
        self.title = title

    @property
    def path(self):  # noqa: D102
        return ""


@dataclass
class ChartGroup:
    """Represents a group of charts like "Deviation"."""

    name: str
    pages: List[vm.Page]
    incomplete_pages: List[IncompletePage]
    intro_text: str
    icon: str = ""  # ALL_CHART_GROUP is the only one that doesn't require an icon.


deviation_intro_text = """
Deviation enables you to draw attention to variations (+/-) from a fixed reference point.
Often this reference point is zero, but you might also show a target or a long term average.
You can also use deviation to express a positive, neutral or negative sentiment.
"""
deviation_chart_group = ChartGroup(
    name="Deviation",
    pages=pages.deviation.pages,
    incomplete_pages=[
        IncompletePage(title="Diverging bar"),
        IncompletePage("Diverging stacked bar"),
        IncompletePage(title="Surplus deficit filled line"),
    ],
    icon="Contrast Square",
    intro_text=deviation_intro_text,
)


correlation_intro_text = """
Correlation helps you show the relationship between two or more variables. It is important that you
make it clear to your audience whether or not the relationship is causal, i.e., whether one causes the
other.
"""
correlation_chart_group = ChartGroup(
    name="Correlation",
    pages=pages.correlation.pages,
    incomplete_pages=[
        IncompletePage("Heatmap matrix"),
    ],
    icon="Bubble Chart",
    intro_text=correlation_intro_text,
)


ranking_intro_text = """
Ranking enables you to present items in an ordered list. Use this when you want to highlight the
position of an item rather than its absolute or relative value. You might want to emphasize the most
interesting points with highlighting or labels to ensure the reader understands what matters most.
"""
ranking_chart_group = ChartGroup(
    name="Ranking",
    pages=pages.ranking.pages,
    incomplete_pages=[
        IncompletePage("Ordered bubble"),
        IncompletePage("Slope"),
        IncompletePage("Lollipop"),
        IncompletePage("Bump"),
    ],
    icon="Stacked Bar Chart",
    intro_text=ranking_intro_text,
)


distribution_intro_text = """
Distribution helps you to present all the possible values (or intervals) of your data and how often they
occur. You can organize the data to show the number or percentage of items in a specified group, what shape
the group takes, where the center lies, and how much variability there is in the data. This shape
(or _skew_) of a distribution can be a powerful way for you to highlight either the existence or lack of
uniformity or equality in the data.
"""
distribution_chart_group = ChartGroup(
    name="Distribution",
    pages=pages.distribution.pages,
    incomplete_pages=[
        IncompletePage("Dot plot"),
        IncompletePage("Barcode"),
        IncompletePage("Cumulative curve"),
        IncompletePage("Beeswarm"),
    ],
    icon="Waterfall Chart",
    intro_text=distribution_intro_text,
)

magnitude_intro_text = """
Magnitude allows you to emphasize size comparisons of **counted** items in your data set. You can show
relative comparisons (whether something is larger or smaller) or absolute differences (where the nuances
are most interesting). Typically, you will use magnitude for actual numbers versus calculated rates or
percentages.
"""
magnitude_chart_group = ChartGroup(
    name="Magnitude",
    pages=pages.magnitude.pages,
    incomplete_pages=[
        IncompletePage("Marimekko"),
        IncompletePage("Lollipop"),
        IncompletePage("Radar"),
        IncompletePage("Pictogram"),
        IncompletePage("Bullet"),
        IncompletePage("Radial"),
    ],
    icon="Bar Chart",
    intro_text=magnitude_intro_text,
)

time_intro_text = """
Time helps you draw attention to important trends emerging over a specified period. The time period you
select could be as short as seconds or as long as centuries. What matters most is selecting the correct
period of time to best show your audience the message they need to take away.
"""
time_chart_group = ChartGroup(
    name="Time",
    pages=pages.time.pages,
    incomplete_pages=[
        IncompletePage("Gantt"),
        IncompletePage("Slope"),
        IncompletePage("Fan"),
        IncompletePage("Bubble timeline"),
        IncompletePage("Sparkline"),
    ],
    icon="Timeline",
    intro_text=time_intro_text,
)


part_to_whole_intro_text = """
Part-to-whole helps you show how one whole item breaks down into its component parts. If you consider the
size of the parts to be most important, a magnitude chart may be more appropriate.
"""
part_to_whole_chart_group = ChartGroup(
    name="Part-to-whole",
    pages=pages.part_to_whole.pages,
    incomplete_pages=[
        IncompletePage("Marimekko"),
        IncompletePage("Arc"),
        IncompletePage("Gridplot"),
        IncompletePage("Venn"),
        IncompletePage("Waterfall"),
    ],
    icon="Donut Small",
    intro_text=part_to_whole_intro_text,
)

flow_intro_text = """
With flow charts, you can highlight the quantity or the intensity of the movement between more than one
state or condition. The flow might be steps in a logical sequence or movement between different geographical
locations.
"""
flow_chart_group = ChartGroup(
    name="Flow",
    pages=pages.flow.pages,
    incomplete_pages=[
        IncompletePage("Waterfall"),
        IncompletePage("Chord"),
        IncompletePage("Network"),
    ],
    icon="Air",
    intro_text=flow_intro_text,
)

spatial_intro_text = """
Spatial charts allow you to demonstrate precise locations or geographical patterns in your data.
"""
spatial_chart_group = ChartGroup(
    name="Spatial",
    pages=pages.spatial.pages,
    incomplete_pages=[
        IncompletePage("Dot map"),
        IncompletePage("Flow map"),
        IncompletePage("Bubble map"),
    ],
    icon="Map",
    intro_text=spatial_intro_text,
)


CHART_GROUPS = [
    deviation_chart_group,
    correlation_chart_group,
    ranking_chart_group,
    distribution_chart_group,
    magnitude_chart_group,
    time_chart_group,
    part_to_whole_chart_group,
    flow_chart_group,
    spatial_chart_group,
]

all_intro_text = """
This dashboard shows a gallery of charts. It includes guidance on when to use each chart type and sample Python code
to create them using [Plotly](https://plotly.com/python/) and [Vizro](https://github.com/mckinsey/vizro).

&nbsp;

&nbsp;

&nbsp;

**Created by:**
- [Huong Li Nguyen](https://github.com/huong-li-nguyen) and [Antony Milne](https://github.com/antonymilne)

- Images created by QuantumBlack

**Inspired by:**
- [The FT Visual Vocabulary](https://github.com/Financial-Times/chart-doctor/blob/main/visual-vocabulary/README.md):
Alan Smith, Chris Campbell, Ian Bott, Liz Faunce, Graham Parrish, Billy Ehrenberg, Paul McCallum,Martin Stabe.

- [The Graphic Continuum](https://www.informationisbeautifulawards.com/showcase/611-the-graphic-continuum):
Jon Swabish and Severino Ribecca

&nbsp;

**Credits and sources:**

- [Plotly](https://plotly.com/python/plotly-express/)
- [Guide to data chart mastery](https://www.atlassian.com/data/charts)

"""


# This contains all pages used across all chart groups, without de-duplicating. De-duplication is done where required
# by remove_duplicates.
ALL_CHART_GROUP = ChartGroup(
    name="All",
    pages=list(itertools.chain(*(chart_group.pages for chart_group in CHART_GROUPS))),
    incomplete_pages=list(itertools.chain(*(chart_group.incomplete_pages for chart_group in CHART_GROUPS))),
    intro_text=all_intro_text,
)



"""Unit tests for vizro.models.Figure."""

import re

import pytest
from asserts import assert_component_equal
from dash import dcc, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
from vizro.figures import kpi_card
from vizro.managers import data_manager


@pytest.fixture
def kpi_card_with_str_dataframe():
    return kpi_card(
        data_frame="gapminder",
        value_column="lifeExp",
        agg_func="mean",
        value_format="{value:.3f}",
    )


class TestFigureInstantiation:
    def test_create_figure_mandatory_only(self, standard_kpi_card):
        figure = vm.Figure(figure=standard_kpi_card)

        assert hasattr(figure, "id")
        assert figure.type == "figure"
        assert figure.figure == standard_kpi_card

    def test_captured_callable_invalid(self, standard_go_chart):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "Invalid CapturedCallable. Supply a function imported from vizro.figures or "
                "defined with decorator @capture('figure')."
            ),
        ):
            vm.Figure(figure=standard_go_chart)

    def test_captured_callable_wrong_mode(self, standard_dash_table):
        with pytest.raises(
            ValidationError,
            match=re.escape(
                "CapturedCallable was defined with @capture('table') rather than @capture('figure') and so "
                "is not compatible with the model."
            ),
        ):
            vm.Figure(figure=standard_dash_table)

    def test_is_model_inheritable(self, standard_kpi_card):
        class MyFigure(vm.Figure):
            pass

        my_figure = MyFigure(figure=standard_kpi_card)

        assert hasattr(my_figure, "id")
        assert my_figure.type == "figure"
        assert my_figure.figure == standard_kpi_card


class TestDunderMethodsFigure:
    def test_getitem_known_args(self, standard_kpi_card):
        figure = vm.Figure(figure=standard_kpi_card)
        assert figure["value_column"] == "lifeExp"
        assert figure["agg_func"] == "mean"
        assert figure["value_format"] == "{value:.3f}"
        assert figure["type"] == "figure"

    def test_getitem_unknown_args(self, standard_kpi_card):
        figure = vm.Figure(figure=standard_kpi_card)
        with pytest.raises(KeyError):
            figure["unknown_args"]


class TestProcessFigureDataFrame:
    def test_process_figure_data_frame_str_df(self, kpi_card_with_str_dataframe, gapminder):
        data_manager["gapminder"] = gapminder
        figure = vm.Figure(figure=kpi_card_with_str_dataframe)
        assert data_manager[figure["data_frame"]].load().equals(gapminder)

    def test_process_figure_data_frame_df(self, standard_kpi_card, gapminder):
        figure = vm.Figure(figure=standard_kpi_card)
        assert data_manager[figure["data_frame"]].load().equals(gapminder)


class TestFigureBuild:
    def test_figure_build(self, standard_kpi_card, gapminder):
        figure = vm.Figure(id="figure-id", figure=standard_kpi_card).build()

        expected_figure = dcc.Loading(
            html.Div(
                id="figure-id",
                className="figure-container",
            ),
            color="grey",
            parent_className="loading-container",
            overlay_style={"visibility": "visible", "opacity": 0.3},
        )
        assert_component_equal(figure, expected_figure)



"""Contains utilities to extract the Action and ActionsChain models from registered pages only."""

from __future__ import annotations

from typing import TYPE_CHECKING, List

import dash

from vizro.managers import model_manager
from vizro.managers._model_manager import ModelID

if TYPE_CHECKING:
    from vizro.models import Action, Page
    from vizro.models._action._actions_chain import ActionsChain


def _get_actions_chains_on_all_pages() -> List[ActionsChain]:
    """Gets list of ActionsChain models for registered pages."""
    actions_chains: List[ActionsChain] = []
    # TODO: once dash.page_registry matches up with model_manager, change this to use purely model_manager.
    # Making the change now leads to problems since there can be Action models defined that aren't used in the
    # dashboard.
    # See https://github.com/mckinsey/vizro/pull/366.
    for registered_page in dash.page_registry.values():
        try:
            page: Page = model_manager[registered_page["module"]]
        except KeyError:
            continue
        actions_chains.extend(model_manager._get_page_actions_chains(page_id=ModelID(str(page.id))))
    return actions_chains


def _get_actions_on_registered_pages() -> List[Action]:
    """Gets list of Action models for registered pages."""
    return [action for action_chain in _get_actions_chains_on_all_pages() for action in action_chain.actions]



"""Example to show dashboard configuration specified as a YAML file."""

from pathlib import Path

import pandas as pd
import vizro.plotly.express as px
import yaml
from vizro import Vizro
from vizro.managers import data_manager
from vizro.models import Dashboard

data_manager["iris"] = px.data.iris()
data_manager["gapminder"] = px.data.gapminder()
data_manager["gapminder_2007"] = px.data.gapminder().query("year == 2007")
data_manager["tips"] = px.data.tips()

df_stocks_long = pd.melt(
    px.data.stocks(datetimes=True),
    id_vars="date",
    value_vars=["GOOG", "AAPL", "AMZN", "FB", "NFLX", "MSFT"],
    var_name="stocks",
    value_name="value",
)
data_manager["df_stocks_long"] = df_stocks_long

df_kpi = pd.DataFrame({"Actual": [100, 200, 700], "Reference": [100, 300, 500], "Category": ["A", "B", "C"]})
data_manager["df_kpi"] = df_kpi

dashboard = yaml.safe_load(Path("dashboard.yaml").read_text(encoding="utf-8"))
dashboard = Dashboard(**dashboard)

if __name__ == "__main__":
    Vizro(assets_folder="../assets").build(dashboard).run()



"""Flow charts."""

import vizro.models as vm
from custom_charts import sankey

from pages._pages_utils import PAGE_GRID, make_code_clipboard_from_py_file, sankey_data

sankey = vm.Page(
    title="Sankey",
    path="flow/sankey",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
            #### What is a sankey chart?

            A Sankey chart is a type of flow diagram that illustrates how resources or values move between different
            stages or entities. The width of the arrows in the chart is proportional to the quantity of the flow,
            making it easy to see where the largest movements occur.

            &nbsp;

            #### When should I use it?

            Use a Sankey chart when you want to visualize the flow of resources, energy, money, or other values from
            one point to another. It is particularly useful for showing distributions and transfers within a system,
            such as energy usage, cost breakdowns, or material flows.

            Be mindful that Sankey charts can become cluttered if there are too many nodes or flows.
            To maintain clarity, focus on highlighting the most significant flows and keep the chart as simple as
            possible.
        """
        ),
        vm.Graph(
            figure=sankey(
                sankey_data,
                labels=["A1", "A2", "B1", "B2", "C1", "C2"],
                source="Origin",
                target="Destination",
                value="Value",
            ),
        ),
        make_code_clipboard_from_py_file("sankey.py"),
    ],
)

pages = [sankey]



"""Unit tests for TextArea."""

import dash_bootstrap_components as dbc
from asserts import assert_component_equal
from dash import html
from vizro.models._components.form._text_area import TextArea


class TestTextAreaInstantiation:
    """Tests model instantiation."""

    def test_create_text_area_mandatory_only(self):
        text_area = TextArea()

        assert hasattr(text_area, "id")
        assert text_area.type == "text_area"
        assert text_area.title == ""
        assert text_area.placeholder == ""
        assert text_area.actions == []

    def test_create_text_area_mandatory_and_optional(self):
        text_area = TextArea(id="text-area-id", title="Title", placeholder="Placeholder")

        assert text_area.id == "text-area-id"
        assert text_area.type == "text_area"
        assert text_area.title == "Title"
        assert text_area.placeholder == "Placeholder"
        assert text_area.actions == []


class TestUserInputBuild:
    """Tests model build method."""

    def test_text_area_build(self):
        text_area = TextArea(id="text-area-id", title="Title", placeholder="Placeholder").build()
        expected_text_area = html.Div(
            [
                dbc.Label("Title", html_for="text-area-id"),
                dbc.Textarea(
                    id="text-area-id",
                    placeholder="Placeholder",
                    persistence=True,
                    persistence_type="session",
                    debounce=True,
                ),
            ]
        )
        assert_component_equal(text_area, expected_text_area)



"""Time charts."""

import vizro.models as vm
import vizro.plotly.express as px
from custom_charts import categorical_column

from pages._factories import column_and_line_factory, connected_scatter_factory
from pages._pages_utils import PAGE_GRID, gapminder, make_code_clipboard_from_py_file, stepped_line_data, stocks, tips

line = vm.Page(
    title="Line",
    path="time/line",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a line chart?

            A line chart presents a series of data points over a continuous interval or time period, joined together
            with straight lines.

            &nbsp;

            #### When should I use it?

            You should select a line chart when you want to show trends over time. Usually, your y-axis will show a
            quantitative value and your x-axis will be marked as a timescale or a sequence of intervals. You can also
            display negative values below the x-axis. If you wish to group several lines (different data series) in the
            same chart, try to limit yourself to 3-4 to avoid cluttering up your chart.
        """
        ),
        vm.Graph(figure=px.line(stocks, x="date", y="GOOG")),
        make_code_clipboard_from_py_file("line.py"),
    ],
)

# Note: Code example for magnitude/column differs from time/column. The text description is the same.
column = vm.Page(
    id="time-column",
    path="time/column",
    title="Column",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
                #### What is a column chart?

                A column chart is a type of bar chart where data is represented with vertical columns. Each
                column's height corresponds to the value it represents, with the y-axis starting from zero.

                &nbsp;

                #### When should I use it?

                Use a column chart to compare sizes and identify patterns in categorical data, including time-based
                data. Arrange columns to fit your message, and for time-based data, order them chronologically to
                highlight trends. Ensure clear labeling, especially with many categories, and consider using a legend
                or abbreviations with fuller descriptions below.
        """
        ),
        vm.Graph(
            figure=categorical_column(
                gapminder.query("country == 'Nigeria' and year > 1970"),
                y="gdpPercap",
                x="year",
            )
        ),
        make_code_clipboard_from_py_file("time_column.py"),
    ],
)

area = vm.Page(
    title="Area",
    path="time/area",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is an area chart?

            An area chart displays data points over a continuous interval or time period, with the area between the line
            and the axis filled in to emphasize the magnitude of the values.

            &nbsp;

            #### When should I use it?

            An area chart is ideal for showing trends over time and emphasizing the volume of data. Typically,
            the y-axis represents a quantitative value, while the x-axis is marked with a timescale or sequence of
            intervals. Area charts can also display negative values below the x-axis. If you need to compare multiple
            data series in the same chart, try to limit yourself to 3-4 to maintain clarity and avoid clutter.
        """
        ),
        vm.Graph(figure=px.area(stocks, x="date", y="GOOG")),
        make_code_clipboard_from_py_file("area.py"),
    ],
)

connected_scatter = connected_scatter_factory("time")
column_and_line = column_and_line_factory("time")

stepped_line = vm.Page(
    title="Stepped line",
    path="time/stepped-line",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""

            #### What is a stepped line chart?

            A stepped line chart is much like a standard line chart but, instead of connecting two points with the
            shortest line, the line forms a series of steps between data points.

            &nbsp;

            #### When should I use it?

            You should use a stepped line chart when you wish to draw attention to changes occurring at specific points.
            By contrast, a line chart would suggest that changes occur gradually.
        """
        ),
        vm.Graph(
            figure=px.line(
                data_frame=stepped_line_data,
                x="year",
                y="rate",
                line_shape="vh",
            ),
        ),
        make_code_clipboard_from_py_file("stepped_line.py"),
    ],
)

heatmap = vm.Page(
    title="Heatmap",
    path="time/heatmap",
    layout=vm.Layout(grid=PAGE_GRID),
    components=[
        vm.Card(
            text="""
            #### What is a heatmap?
            A heatmap chart depicts values for a main variable of interest across two axis variables as a grid of
            colored squares. The color intensity of each cell represents the value of the main variable within a
            specific range.

            &nbsp;

            #### When should I use it?

            Use a heatmap chart to visualize time patterns and identify trends between two variables.
            Typically, the x-axis represents time intervals (e.g., hours, days, months), while the y-axis represents
            categories or different variables. By observing color changes across the grid, you can easily spot
            patterns and correlations.
        """
        ),
        vm.Graph(figure=px.density_heatmap(tips, x="day", y="size", z="tip", histfunc="avg", text_auto="$.2f")),
        make_code_clipboard_from_py_file("heatmap.py"),
    ],
)
pages = [line, column, area, connected_scatter, column_and_line, stepped_line, heatmap]



import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro

gapminder = px.data.gapminder()

page = vm.Page(
    title="Connected scatter",
    components=[
        vm.Graph(figure=px.line(gapminder.query("country == 'Australia'"), x="year", y="lifeExp", markers=True))
    ],
)

dashboard = vm.Dashboard(pages=[page])
Vizro().build(dashboard).run()



"""Contains unwrapped KPI card functions (suitable to use in pure Dash app)."""

from vizro.figures import kpi_card, kpi_card_reference

kpi_card = kpi_card.__wrapped__
kpi_card_reference = kpi_card_reference.__wrapped__

__all__ = ["kpi_card", "kpi_card_reference"]



"""Types used in pydantic fields."""

# ruff: noqa: F821
from __future__ import annotations

import functools
import importlib
import inspect
from contextlib import contextmanager
from datetime import date
from typing import Any, Dict, List, Literal, Protocol, Union, runtime_checkable

import plotly.io as pio

try:
    from pydantic.v1 import Field, StrictBool
    from pydantic.v1.fields import ModelField
    from pydantic.v1.schema import SkipField
except ImportError:  # pragma: no cov
    from pydantic import Field, StrictBool
    from pydantic.fields import ModelField
    from pydantic.schema import SkipField


from typing_extensions import Annotated, TypedDict

from vizro.charts._charts_utils import _DashboardReadyFigure


def _clean_module_string(module_string: str) -> str:
    from vizro.models._models_utils import REPLACEMENT_STRINGS

    for original, new in REPLACEMENT_STRINGS.items():
        if original in module_string:
            return new
    return ""


# Used to describe _DashboardReadyFigure, so we can keep CapturedCallable generic rather than referring to
# _DashboardReadyFigure explicitly.
@runtime_checkable
class _SupportsCapturedCallable(Protocol):
    _captured_callable: CapturedCallable


class CapturedCallable:
    """Stores a captured function call to use in a dashboard.

    Users do not need to instantiate this class directly. Instances are instead generated automatically
    through the [`capture`][vizro.models.types.capture] decorator. Some of the functionality is similar to
    `functools.partial`.

    Ready-to-use `CapturedCallable` instances are provided by Vizro. In this case refer to the [user guide on
    Charts/Graph](../user-guides/graph.md), [Table](../user-guides/table.md), [Actions](../user-guides/actions.md)
    or [Figures](../user-guides/figure.md) to see available choices.

    (Advanced) In case you would like to create your own `CapturedCallable`, please refer to the [user guide on
    custom charts](../user-guides/custom-charts.md),
    [custom tables](../user-guides/custom-tables.md),
    [custom actions](../user-guides/custom-actions.md),
    or [custom figures](../user-guides/custom-figures.md).
    """

    def __init__(self, function, /, *args, **kwargs):
        """Creates a new `CapturedCallable` object that will be able to re-run `function`.

        Partially binds *args and **kwargs to the function call.

        Raises
            ValueError if `function` contains positional-only or variadic positional parameters (*args).

        """
        # It is difficult to get positional-only and variadic positional arguments working at the same time as
        # variadic keyword arguments. Ideally we would do the __call__ as
        # self.__function(*bound_arguments.args, **bound_arguments.kwargs) as in the
        # Python documentation. This would handle positional-only and variadic positional arguments better but makes
        # it more difficult to handle variadic keyword arguments due to https://bugs.python.org/issue41745.
        # Hence we abandon bound_arguments.args and bound_arguments.kwargs in favor of just using
        # self.__function(**bound_arguments.arguments).
        parameters = inspect.signature(function).parameters
        invalid_params = {
            param.name
            for param in parameters.values()
            if param.kind in [inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.VAR_POSITIONAL]
        }

        if invalid_params:
            raise ValueError(
                f"Invalid parameter {', '.join(invalid_params)}. CapturedCallable does not accept functions with "
                f"positional-only or variadic positional parameters (*args)."
            )

        self.__function = function
        self.__bound_arguments = inspect.signature(function).bind_partial(*args, **kwargs).arguments
        self.__unbound_arguments = [
            param for param in parameters.values() if param.name not in self.__bound_arguments
        ]  # Maintaining the same order here is important.

        # A function can only ever have one variadic keyword parameter. {""} is just here so that var_keyword_param
        # is always unpacking a one element set.
        (var_keyword_param,) = {
            param.name for param in parameters.values() if param.kind == inspect.Parameter.VAR_KEYWORD
        } or {""}

        # Since we do __call__ as self.__function(**bound_arguments.arguments), we need to restructure the arguments
        # a bit to put the kwargs in the right place.
        # For a function with parameter **kwargs this converts self.__bound_arguments = {"kwargs": {"a": 1}} into
        # self.__bound_arguments = {"a": 1}.
        if var_keyword_param in self.__bound_arguments:
            self.__bound_arguments.update(self.__bound_arguments[var_keyword_param])
            del self.__bound_arguments[var_keyword_param]

        # Used in later validations of the captured callable.
        self._mode = None
        self._model_example = None

    def __call__(self, *args, **kwargs):
        """Run the `function` with the initially bound arguments overridden by `**kwargs`.

        *args are possible here, but cannot be used to override arguments bound in `__init__` - just to
        provide additional arguments. You can still override arguments that were originally given
        as positional using their argument name.
        """
        if args and kwargs:
            # In theory we could probably lift this restriction, but currently we don't need to and we'd need
            # to give careful thought on the right way to handle cases where there's ambiguity in the
            # self.__function call as the same argument is potentially being provided through both *args and **kwargs.
            raise ValueError("CapturedCallable does not support calling with both positional and keyword arguments.")

        # In order to avoid any ambiguity in the call to self.__function, we cannot provide use the *args directly.
        # Instead they must converted to keyword arguments and so we need to match them up with the right keywords.
        # Since positional-only or variadic positional parameters are not possible (they raise ValueError in __init__)
        # the only possible type of argument *args could be address is positional-or-keyword.
        if args:
            unbound_positional_arguments = [
                param.name
                for param in self.__unbound_arguments
                if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            ]
            if len(args) > len(unbound_positional_arguments):
                # TypeError to match the standard Python exception raised in this case.
                raise TypeError(
                    f"CapturedCallable takes {len(unbound_positional_arguments)} "
                    f"positional arguments but {len(args)} were given."
                )

            # No need to handle case that len(args) < len(unbound_positional_arguments),
            # since this will already raise error in the following function call.
            return self.__function(**dict(zip(unbound_positional_arguments, args)), **self.__bound_arguments)

        return self.__function(**{**self.__bound_arguments, **kwargs})

    def __getitem__(self, arg_name: str):
        """Gets the value of a bound argument."""
        return self.__bound_arguments[arg_name]

    def __setitem__(self, arg_name: str, value):
        """Sets the value of a bound argument."""
        self.__bound_arguments[arg_name] = value

    @property
    def _arguments(self):
        # TODO: This is used twice: in _get_parametrized_config and in vm.Action and should be removed when those
        # references are removed.
        return self.__bound_arguments

    # TODO-actions: Find a way how to compare CapturedCallable and function
    @property
    def _function(self):
        return self.__function

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any], field: ModelField):
        """Generates schema for field of this type."""
        raise SkipField(f"{cls.__name__} {field.name} is excluded from the schema.")

    @classmethod
    def __get_validators__(cls):
        """Makes type compatible with pydantic model without needing `arbitrary_types_allowed`."""
        # Each validator receives as an input the value returned from the previous validator.
        # captured_callable could be _SupportsCapturedCallable, CapturedCallable, dictionary from JSON/YAML or
        # invalid type at this point. Begin by parsing it from JSON/YAML if applicable:
        yield cls._parse_json
        # At this point captured_callable is _SupportsCapturedCallable, CapturedCallable or invalid type. Next extract
        # it from _SupportsCapturedCallable if applicable:
        yield cls._extract_from_attribute
        # At this point captured_callable is CapturedCallable or invalid type. Check it is in fact CapturedCallable
        # and do final checks:
        yield cls._check_type

    @classmethod
    def _parse_json(
        cls,
        captured_callable_config: Union[_SupportsCapturedCallable, CapturedCallable, Dict[str, Any]],
        field: ModelField,
    ) -> Union[CapturedCallable, _SupportsCapturedCallable]:
        """Parses captured_callable_config specification from JSON/YAML.

        If captured_callable_config is already _SupportCapturedCallable or CapturedCallable then it just passes through
        untouched.

        This uses the hydra syntax for _target_ but none of the other bits and we don't actually use hydra
        to implement it. In future, we might like to switch to using hydra's actual implementation
        which would allow nested functions (e.g. for transformers?) and to specify the path to a _target_ that lives
        outside of vizro.plotly_express. See https://hydra.cc/docs/advanced/instantiate_objects/overview/.
        """
        if not isinstance(captured_callable_config, dict):
            return captured_callable_config

        # Try to import function given in _target_ from the import_path property of the pydantic field.
        try:
            function_name = captured_callable_config.pop("_target_")
        except KeyError as exc:
            raise ValueError(
                "CapturedCallable object must contain the key '_target_' that gives the target function."
            ) from exc

        import_path = field.field_info.extra["import_path"]
        try:
            function = getattr(importlib.import_module(import_path), function_name)
        except (AttributeError, ModuleNotFoundError) as exc:
            raise ValueError(f"_target_={function_name} cannot be imported from {import_path}.") from exc

        # All the other items in figure are the keyword arguments to pass into function.
        function_kwargs = captured_callable_config

        # It would seem natural to return cls(function, **function_kwargs) here, but the function is already decorated
        # with @capture, and so that would return a nested CapturedCallable.
        return function(**function_kwargs)

    @classmethod
    def _extract_from_attribute(
        cls, captured_callable: Union[_SupportsCapturedCallable, CapturedCallable]
    ) -> CapturedCallable:
        """Extracts CapturedCallable from _SupportCapturedCallable (e.g. _DashboardReadyFigure).

        If captured_callable is already CapturedCallable then it just passes through untouched.
        """
        if not isinstance(captured_callable, _SupportsCapturedCallable):
            return captured_callable
        return captured_callable._captured_callable

    @classmethod
    def _check_type(cls, captured_callable: CapturedCallable, field: ModelField) -> CapturedCallable:
        """Checks captured_callable is right type and mode."""
        expected_mode = field.field_info.extra["mode"]
        import_path = field.field_info.extra["import_path"]

        if not isinstance(captured_callable, CapturedCallable):
            raise ValueError(
                f"Invalid CapturedCallable. Supply a function imported from {import_path} or defined with "
                f"decorator @capture('{expected_mode}')."
            )

        if (mode := captured_callable._mode) != expected_mode:
            raise ValueError(
                f"CapturedCallable was defined with @capture('{mode}') rather than @capture('{expected_mode}') and so "
                "is not compatible with the model."
            )

        return captured_callable

    def __repr__(self):
        """String representation of the CapturedCallable."""
        args = ", ".join(f"{key}={value!r}" for key, value in self._arguments.items())
        return f"{self._function.__module__}.{self._function.__name__}({args})"

    def __repr_clean__(self):
        """Alternative __repr__ method with cleaned module paths."""
        args = ", ".join(f"{key}={value!r}" for key, value in self._arguments.items())
        original_module_path = f"{self._function.__module__}"
        return f"{_clean_module_string(original_module_path)}{self._function.__name__}({args})"


@contextmanager
def _pio_templates_default():
    """Sets pio.templates.default to "vizro_dark" and then reverts it.

    This is to ensure that in a Jupyter notebook captured charts look the same as when they're in the dashboard. When
    the context manager exits the global theme is reverted just to keep things clean (e.g. if you really wanted to,
    you could compare a captured vs. non-captured chart in the same Python session).

    This works even if users have tweaked the templates, so long as pio.templates has been updated correctly and you
    refer to template by name rather than trying to take from vizro.themes.

    If pio.templates.default has already been set to vizro_dark or vizro_light then no change is made to allow a user
    to set these without it being overridden.
    """
    old_default = pio.templates.default
    template_changed = False
    # If the user has set pio.templates.default to a vizro theme already, no need to change it.
    if old_default not in ["vizro_dark", "vizro_light"]:
        template_changed = True
        pio.templates.default = "vizro_dark"

    # Revert the template. This is done in a try/finally so that if the code wrapped inside the context manager (i.e.
    # plotting functions) raises an exception, pio.templates.default is still reverted. This is not very important
    # but easy to achieve.
    try:
        # This will always be vizro_light or vizro_dark and corresponds to the default theme that has been set.
        yield pio.templates.default
    finally:
        if template_changed:
            pio.templates.default = old_default


class capture:
    """Captures a function call to create a [`CapturedCallable`][vizro.models.types.CapturedCallable].

    This is used to add the functionality required to make graphs and actions work in a dashboard.
    Typically, it should be used as a function decorator. There are five possible modes: `"graph"`, `"table"`,
    `"ag_grid"`, `"figure"` and `"action"`.

    Examples
        >>> @capture("graph")
        >>> def graph_function():
        >>>     ...
        >>> @capture("table")
        >>> def table_function():
        >>>     ...
        >>> @capture("ag_grid")
        >>> def ag_grid_function():
        >>>     ...
        >>> @capture("figure")
        >>> def figure_function():
        >>>     ...
        >>> @capture("action")
        >>> def action_function():
        >>>     ...

    For further help on the use of `@capture("graph")`, you can refer to the guide on
    [custom graphs](../user-guides/custom-charts.md).
    For further help on the use of `@capture("table")` or `@capture("ag_grid")`, you can refer to the guide on
    [custom tables](../user-guides/custom-tables.md).
    For further help on the use of `@capture("figure")`, you can refer to the guide on
    [figures](../user-guides/figure.md).
    For further help on the use of `@capture("action")`, you can refer to the guide on
    [custom actions](../user-guides/custom-actions.md).

    """

    def __init__(self, mode: Literal["graph", "action", "table", "ag_grid", "figure"]):
        """Decorator to capture a function call."""
        # mode and model_example are used in later validations of the captured callable.
        self._mode = mode
        model_examples = {
            "graph": "vm.Graph(figure=...)",
            "action": "vm.Action(function=...)",
            "table": "vm.Table(figure=...)",
            "ag_grid": "vm.AgGrid(figure=...)",
            "figure": "vm.Figure(figure=...)",
        }
        self._model_example = model_examples[mode]

    def __call__(self, func, /):
        """Produces a CapturedCallable or _DashboardReadyFigure.

        mode="action" and mode="table" give a CapturedCallable, while mode="graph" gives a _DashboardReadyFigure that
        contains a CapturedCallable. In both cases, the CapturedCallable is based on func and the provided
        *args and **kwargs.
        """
        if self._mode == "graph":
            # The more difficult case, where we need to still have a valid plotly figure that renders in a notebook.
            # Hence we attach the CapturedCallable as a property instead of returning it directly.
            # TODO: move point of checking that data_frame argument exists earlier on.
            # TODO: also would be nice to raise errors in CapturedCallable.__init__ at point of function definition
            #  rather than point of calling if possible.
            @functools.wraps(func)
            def wrapped(*args, **kwargs) -> _DashboardReadyFigure:
                if "data_frame" not in inspect.signature(func).parameters:
                    raise ValueError(f"{func.__name__} must have data_frame argument to use capture('graph').")

                # We need to capture function upfront in order to find value of data_frame argument: since it could be
                # positional or keyword, this is much more robust than trying to get it out of arg or kwargs ourselves.
                captured_callable: CapturedCallable = CapturedCallable(func, *args, **kwargs)
                captured_callable._mode = self._mode
                captured_callable._model_example = self._model_example

                try:
                    captured_callable["data_frame"]
                except KeyError as exc:
                    raise ValueError(f"{func.__name__} must supply a value to data_frame argument.") from exc

                if isinstance(captured_callable["data_frame"], str):
                    # Enable running e.g. px.scatter("iris") from the Python API. Don't actually run the function
                    # because it won't work as there's no data. This case is not relevant for the JSON/YAML API,
                    # which is handled separately through validation of CapturedCallable.
                    fig = _DashboardReadyFigure()
                else:
                    # Standard case for px.scatter(df: pd.DataFrame).
                    # Set theme for the figure that gets shown in a Jupyter notebook. This is to ensure that in a
                    # Jupyter notebook captured charts look the same as when they're in the dashboard. To mimic this,
                    # we first use _pio_templates_default to set the global theme, as is done in the dashboard, and then
                    # do the fig.layout.template update that is achieved by the theme selector.
                    # We don't want to update the captured_callable in the same way, since it's only used inside the
                    # dashboard, at which point the global pio.templates.default is always set anyway according to
                    # the dashboard theme and then updated according to the theme selector.
                    with _pio_templates_default() as default_template:
                        fig = func(*args, **kwargs)
                    # Update the fig.layout.template just to ensure absolute consistency with how the dashboard
                    # works.
                    # The only exception here is the edge case that the user has specified template="vizro_light" or
                    # "vizro_dark" in the plotting function, in which case we don't want to change it. This makes
                    # it easier for a user to try out both themes simultaneously in a notebook.
                    if fig.layout.template not in (pio.templates["vizro_dark"], pio.templates["vizro_light"]):
                        fig.layout.template = default_template
                    fig.__class__ = _DashboardReadyFigure

                fig._captured_callable = captured_callable
                return fig

            return wrapped
        elif self._mode == "action":
            # The "normal" case where we just capture the function call.
            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                # Note this is basically the same as partial(func, *args, **kwargs)
                captured_callable: CapturedCallable = CapturedCallable(func, *args, **kwargs)
                captured_callable._mode = self._mode
                captured_callable._model_example = self._model_example
                return captured_callable

            return wrapped
        elif self._mode in ["table", "ag_grid", "figure"]:

            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                if "data_frame" not in inspect.signature(func).parameters:
                    raise ValueError(f"{func.__name__} must have data_frame argument to use capture('table').")

                captured_callable: CapturedCallable = CapturedCallable(func, *args, **kwargs)
                captured_callable._mode = self._mode
                captured_callable._model_example = self._model_example

                try:
                    captured_callable["data_frame"]
                except KeyError as exc:
                    raise ValueError(f"{func.__name__} must supply a value to data_frame argument.") from exc
                return captured_callable

            return wrapped
        raise ValueError(
            "Valid modes of the capture decorator are @capture('graph'), @capture('action'), @capture('table'), "
            "@capture('ag_grid') and @capture('figure')."
        )


# Types used for selector values and options. Note the docstrings here are rendered on the API reference.
SingleValueType = Union[StrictBool, float, str, date]
"""Permissible value types for single-value selectors. Values are displayed as default."""
MultiValueType = Union[List[StrictBool], List[float], List[str], List[date]]
"""Permissible value types for multi-value selectors. Values are displayed as default."""


class OptionsDictType(TypedDict):
    """Permissible sub-type for OptionsType. Needs to be in the format of {"label": XXX, "value": XXX}."""

    label: str
    value: SingleValueType


OptionsType = Union[List[StrictBool], List[float], List[str], List[date], List[OptionsDictType]]
"""Permissible options types for selectors. Options are available choices for user to select from."""

# All the below types rely on models and so must use ForwardRef (i.e. "Checklist" rather than actual Checklist class).
SelectorType = Annotated[
    Union["Checklist", "DatePicker", "Dropdown", "RadioItems", "RangeSlider", "Slider"],
    Field(discriminator="type", description="Selectors to be used inside a control."),
]
"""Discriminated union. Type of selector to be used inside a control: [`Checklist`][vizro.models.Checklist],
[`DatePicker`][vizro.models.DatePicker], [`Dropdown`][vizro.models.Dropdown], [`RadioItems`][vizro.models.RadioItems],
[`RangeSlider`][vizro.models.RangeSlider] or [`Slider`][vizro.models.Slider]."""

_FormComponentType = Annotated[
    Union[SelectorType, "Button", "UserInput"],
    Field(discriminator="type", description="Components that can be used to receive user input within a form."),
]

ControlType = Annotated[
    Union["Filter", "Parameter"],
    Field(discriminator="type", description="Control that affects components on the page."),
]
"""Discriminated union. Type of control that affects components on the page: [`Filter`][vizro.models.Filter] or
[`Parameter`][vizro.models.Parameter]."""

ComponentType = Annotated[
    Union["AgGrid", "Button", "Card", "Container", "Figure", "Graph", "Table", "Tabs"],
    Field(
        discriminator="type",
        description="Component that makes up part of the layout on the page.",
    ),
]
"""Discriminated union. Type of component that makes up part of the layout on the page:
[`Button`][vizro.models.Button], [`Card`][vizro.models.Card], [`Table`][vizro.models.Table],
[`Graph`][vizro.models.Graph] or [`AgGrid`][vizro.models.AgGrid]."""

NavPagesType = Union[List[str], Dict[str, List[str]]]
"List of page IDs or a mapping from name of a group to a list of page IDs (for hierarchical sub-navigation)."

NavSelectorType = Annotated[
    Union["Accordion", "NavBar"], Field(discriminator="type", description="Component for rendering navigation.")
]
"""Discriminated union. Type of component for rendering navigation:
[`Accordion`][vizro.models.Accordion] or [`NavBar`][vizro.models.NavBar]."""



import pandas as pd
import plotly.graph_objects as go
import vizro.models as vm
from vizro import Vizro
from vizro.models.types import capture

ages = pd.DataFrame(
    {
        "Age": ["0-19", "20-29", "30-39", "40-49", "50-59", ">=60"],
        "Male": [800, 2000, 4200, 5000, 2100, 800],
        "Female": [1000, 3000, 3500, 3800, 3600, 700],
    }
)


@capture("graph")
def butterfly(data_frame: pd.DataFrame, x1: str, x2: str, y: str):
    fig = go.Figure()
    fig.add_trace(
        go.Bar(
            x=-data_frame[x1],
            y=data_frame[y],
            orientation="h",
            name=x1,
        )
    )
    fig.add_trace(
        go.Bar(
            x=data_frame[x2],
            y=data_frame[y],
            orientation="h",
            name=x2,
        )
    )
    fig.update_layout(barmode="relative")
    return fig


dashboard = vm.Dashboard(
    pages=[
        vm.Page(
            title="Butterfly",
            components=[vm.Graph(figure=butterfly(ages, x1="Male", x2="Female", y="Age"))],
        )
    ]
)
Vizro().build(dashboard).run()
